from typing import Optional, List, Dict

from pydantic import BaseModel

from phidata.asset.aws import AwsAsset, AwsAssetArgs
from phidata.infra.aws.resource.s3.bucket import S3Bucket
from phidata.infra.aws.resource.iam.role import IamRole
from phidata.infra.aws.resource.glue.crawler import GlueCrawlerResource
from phidata.utils.cli_console import print_info, print_error
from phidata.utils.log import logger


class GlueS3Target(BaseModel):
    # The directory path in the S3 bucket to target
    dir: str = ""
    # The s3 bucket to target
    bucket: S3Bucket
    # A list of glob patterns used to exclude from the crawl.
    # For more information, see https://docs.aws.amazon.com/glue/latest/dg/add-crawler.html
    exclusions: Optional[List[str]] = None
    # The name of a connection which allows a job or crawler to access data in Amazon S3 within an
    # Amazon Virtual Private Cloud environment (Amazon VPC).
    connection_name: Optional[str] = None
    # Sets the number of files in each leaf folder to be crawled when crawling sample files in a dataset.
    # If not set, all the files are crawled. A valid value is an integer between 1 and 249.
    sample_size: Optional[int] = None
    # A valid Amazon SQS ARN. For example, arn:aws:sqs:region:account:sqs .
    event_queue_arn: Optional[str] = None
    # A valid Amazon dead-letter SQS ARN. For example, arn:aws:sqs:region:account:deadLetterQueue .
    dlq_event_queue_arn: Optional[str] = None


class GlueCrawlerArgs(AwsAssetArgs):
    # Name of the crawler.
    name: str
    # The IAM role for the crawler
    iam_role: IamRole
    # List of GlueS3Target to add to the targets dict
    s3_targets: Optional[List[GlueS3Target]] = None
    # The Glue database where results are written,
    # such as: arn:aws:daylight:us-east-1::database/sometable/* .
    database_name: Optional[str] = None
    # A description of the new crawler.
    description: Optional[str] = None
    # A list of collection of targets to crawl.
    targets: Optional[Dict[str, List[dict]]] = None
    # A cron expression used to specify the schedule
    # For example, to run something every day at 12:15 UTC,
    # you would specify: cron(15 12 * * ? *) .
    schedule: Optional[str] = None
    # A list of custom classifiers that the user has registered.
    # By default, all built-in classifiers are included in a crawl,
    # but these custom classifiers always override the default classifiers for a given classification.
    classifiers: Optional[List[str]] = None
    # The table prefix used for catalog tables that are created.
    table_prefix: Optional[str] = None
    # The policy for the crawler's update and deletion behavior.
    schema_change_policy: Optional[Dict[str, str]] = None
    # A policy that specifies whether to crawl the entire dataset again,
    # or to crawl only folders that were added since the last crawler run.
    recrawl_policy: Optional[Dict[str, str]] = None
    lineage_configuration: Optional[Dict[str, str]] = None
    lake_formation_configuration: Optional[Dict[str, str]] = None
    # Crawler configuration information. This versioned JSON string
    # allows users to specify aspects of a crawler's behavior.
    configuration: Optional[str] = None
    # The name of the SecurityConfiguration structure to be used by this crawler.
    crawler_security_configuration: Optional[str] = None
    # The tags to use with this crawler request.
    tags: Optional[Dict[str, str]] = None

    ## Resource management
    # If True, skip resource creation if active resources with the same name exist.
    use_cache: bool = True
    # If True, logs extra debug messages
    use_verbose_logs: bool = False
    # If True, does not create the resource when `phi ws up` command is run
    skip_create: bool = False
    # If True, does not delete the resource when `phi ws down` command is run
    skip_delete: bool = False


class GlueCrawler(AwsAsset):
    def __init__(
        self,
        # Name of the crawler.
        name: str,
        # The IAM role for the crawler
        iam_role: IamRole,
        # List of GlueS3Target to add to the targets dict
        s3_targets: Optional[List[GlueS3Target]] = None,
        # The Glue database where results are written,
        # such as: arn:aws:daylight:us-east-1::database/sometable/* .
        database_name: Optional[str] = None,
        # A description of the new crawler.
        description: Optional[str] = None,
        # A list of collection of targets to crawl.
        targets: Optional[Dict[str, List[dict]]] = None,
        # A cron expression used to specify the schedule
        # For example, to run something every day at 12:15 UTC,
        # you would specify: cron(15 12 * * ? *) .,
        schedule: Optional[str] = None,
        # A list of custom classifiers that the user has registered.
        # By default, all built-in classifiers are included in a crawl,
        # but these custom classifiers always override the default classifiers for a given classification.
        classifiers: Optional[List[str]] = None,
        # The table prefix used for catalog tables that are created.
        table_prefix: Optional[str] = None,
        # The policy for the crawler's update and deletion behavior.
        schema_change_policy: Optional[Dict[str, str]] = None,
        # A policy that specifies whether to crawl the entire dataset again,
        # or to crawl only folders that were added since the last crawler run.
        recrawl_policy: Optional[Dict[str, str]] = None,
        lineage_configuration: Optional[Dict[str, str]] = None,
        lake_formation_configuration: Optional[Dict[str, str]] = None,
        # Crawler configuration information. This versioned JSON string
        # allows users to specify aspects of a crawler's behavior.
        configuration: Optional[str] = None,
        # The name of the SecurityConfiguration structure to be used by this crawler.
        crawler_security_configuration: Optional[str] = None,
        # The tags to use with this crawler request.
        tags: Optional[Dict[str, str]] = None,
        ## Resource management
        # If True, skip resource creation if active resources with the same name exist.
        use_cache: bool = True,
        # If True, logs extra debug messages
        use_verbose_logs: bool = False,
        # If True, does not create the resource when `phi ws up` command is run
        skip_create: bool = False,
        # If True, does not delete the resource when `phi ws down` command is run
        skip_delete: bool = False,
        version: Optional[str] = None,
        enabled: bool = True,
    ) -> None:

        super().__init__()
        try:
            self.args: GlueCrawlerArgs = GlueCrawlerArgs(
                name=name,
                iam_role=iam_role,
                s3_targets=s3_targets,
                database_name=database_name,
                description=description,
                targets=targets,
                schedule=schedule,
                classifiers=classifiers,
                table_prefix=table_prefix,
                schema_change_policy=schema_change_policy,
                recrawl_policy=recrawl_policy,
                lineage_configuration=lineage_configuration,
                lake_formation_configuration=lake_formation_configuration,
                configuration=configuration,
                crawler_security_configuration=crawler_security_configuration,
                tags=tags,
                use_cache=use_cache,
                use_verbose_logs=use_verbose_logs,
                skip_create=skip_create,
                skip_delete=skip_delete,
                version=version,
                enabled=enabled,
            )
            self._glue_crawler_resource: Optional[GlueCrawlerResource] = None
        except Exception as e:
            print_error(f"Args for {self.__class__.__name__} are not valid")
            raise

    @property
    def glue_crawler_resource(self) -> GlueCrawlerResource:
        # use cached value if available
        if self._glue_crawler_resource:
            return self._glue_crawler_resource

        # Create GlueCrawlerResource
        logger.debug("Initializing GlueCrawlerResource")
        try:
            # Update crawler targets
            # start with user provided targets
            crawler_targets: Optional[Dict[str, List[dict]]] = self.args.targets

            # Add GlueS3Targets to crawler_targets
            if self.args.s3_targets is not None:
                # create S3Targets dicts using s3_targets
                new_s3_targets_list: List[dict] = []
                for s3_target in self.args.s3_targets:
                    _new_s3_target_path = (
                        f"s3://{s3_target.bucket.name}/{s3_target.dir}"
                    )
                    # start with the only required argument
                    _new_s3_target_dict = {"Path": _new_s3_target_path}
                    # add any optional arguments
                    if s3_target.exclusions is not None:
                        _new_s3_target_dict["Exclusions"] = s3_target.exclusions
                    if s3_target.connection_name is not None:
                        _new_s3_target_dict[
                            "ConnectionName"
                        ] = s3_target.connection_name
                    if s3_target.sample_size is not None:
                        _new_s3_target_dict["SampleSize"] = s3_target.sample_size
                    if s3_target.event_queue_arn is not None:
                        _new_s3_target_dict["EventQueueArn"] = s3_target.event_queue_arn
                    if s3_target.dlq_event_queue_arn is not None:
                        _new_s3_target_dict[
                            "DlqEventQueueArn"
                        ] = s3_target.dlq_event_queue_arn

                    new_s3_targets_list.append(_new_s3_target_dict)

                # Add new S3Targets to crawler_targets
                if crawler_targets is None:
                    crawler_targets = {}
                # logger.debug(f"new_s3_targets_list: {new_s3_targets_list}")
                existing_s3_targets = crawler_targets.get("S3Targets", [])
                # logger.debug(f"existing_s3_targets: {existing_s3_targets}")
                new_s3_targets = existing_s3_targets + new_s3_targets_list
                # logger.debug(f"new_s3_targets: {new_s3_targets}")
                crawler_targets["S3Targets"] = new_s3_targets

            # TODO: add more targets as needed
            logger.debug(f"GlueCrawler targets: {crawler_targets}")

            self._glue_crawler_resource = GlueCrawlerResource(
                name=self.args.name,
                iam_role=self.args.iam_role,
                database_name=self.args.database_name,
                description=self.args.description,
                targets=crawler_targets,
                schedule=self.args.schedule,
                classifiers=self.args.classifiers,
                table_prefix=self.args.table_prefix,
                schema_change_policy=self.args.schema_change_policy,
                recrawl_policy=self.args.recrawl_policy,
                lineage_configuration=self.args.lineage_configuration,
                lake_formation_configuration=self.args.lake_formation_configuration,
                configuration=self.args.configuration,
                crawler_security_configuration=self.args.crawler_security_configuration,
                tags=self.args.tags,
                use_cache=self.args.use_cache,
                use_verbose_logs=self.args.use_verbose_logs,
                skip_create=self.args.skip_create,
                skip_delete=self.args.skip_delete,
            )
            return self._glue_crawler_resource
        except Exception as e:
            print_error(f"Could not create GlueCrawlerResource")
            raise

    ######################################################
    ## Create GlueCrawler
    ######################################################

    def create_crawler(self) -> bool:

        # GlueCrawler not yet initialized
        if self.args is None:
            return False

        from phidata.infra.aws.api_client import AwsApiClient

        try:
            # aws_region, aws_profile loaded from local env if provided
            # by WorkspaceConfig
            aws_api_client = AwsApiClient(
                aws_region=self.aws_region,
                aws_profile=self.aws_profile,
            )
            glue_crawler_resource = self.glue_crawler_resource
            create_success = glue_crawler_resource.create(aws_api_client)
            return create_success
        except Exception as e:
            print_error("Could not create GlueCrawler, please try again")
            print_info("--- stacktrace ---")
            logger.exception(e)
            print_info("--- stacktrace ---")
        return False

    ######################################################
    ## Start GlueCrawler
    ######################################################

    def start_crawler(self) -> bool:

        # GlueCrawler not yet initialized
        if self.args is None:
            return False

        from phidata.infra.aws.api_client import AwsApiClient

        try:
            aws_api_client = AwsApiClient(
                aws_region=self.aws_region,
                aws_profile=self.aws_profile,
            )
            glue_crawler_resource = self.glue_crawler_resource
            active_glue_crawler_resource = glue_crawler_resource.read(aws_api_client)
            if active_glue_crawler_resource is None:
                print_error("No GlueCrawler available")
                return False

            start_success = glue_crawler_resource.start_crawler(aws_api_client)
            return start_success
        except Exception as e:
            print_error("Could not start GlueCrawler, please try again")
            print_info("--- stacktrace ---")
            logger.exception(e)
            print_info("--- stacktrace ---")
        return False



from pathlib import Path
from typing import Optional, Any, Dict, List, Literal

from botocore.exceptions import ClientError

from phidata.infra.aws.api_client import AwsApiClient
from phidata.infra.aws.resource.base import AwsResource
from phidata.infra.aws.resource.cloudformation.stack import CloudFormationStack
from phidata.infra.aws.resource.rds.db_subnet_group import DbSubnetGroup
from phidata.utils.cli_console import print_info, print_error, print_warning
from phidata.utils.log import logger


class DbCluster(AwsResource):
    """
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/emr.html
    """

    resource_type = "DbCluster"
    service_name = "rds"

    # Name of the cluster.
    name: str
    # The name of the database engine to be used for this DB cluster.
    engine: Literal["aurora", "aurora-mysql", "aurora-postgresql", "mysql", "postgres"]
    # The compute and memory capacity of each DB instance in the Multi-AZ DB cluster, for example db.m6g.xlarge.
    # Not all DB instance classes are available in all Amazon Web Services Regions, or for all database engines.
    # This setting is required to create a Multi-AZ DB cluster.
    db_instance_class: Optional[str] = None

    # The version number of the database engine to use.
    engine_version: Optional[str] = None
    # The port number on which the instances in the DB cluster accept connections.
    port: Optional[int] = None
    # The name of the master user for the DB cluster.
    # Constraints:
    #   Must be 1 to 16 letters or numbers.
    #   First character must be a letter.
    #   Can't be a reserved word for the chosen database engine.
    # Provide MASTER_USERNAME here or as MASTER_USERNAME in secrets_file
    master_username: Optional[str] = None
    # The password for the master database user. This password can contain any printable ASCII character
    # except "/", """, or "@".
    # Constraints: Must contain from 8 to 41 characters.
    # Provide MASTER_USER_PASSWORD here or as MASTER_USER_PASSWORD in secrets_file
    master_user_password: Optional[str] = None
    # The name for your database of up to 64 alphanumeric characters.
    # If you do not provide a name, Amazon RDS doesn't create a database in the DB cluster you are creating.
    # Provide DATABASE_NAME here or as DATABASE_NAME in secrets_file
    database_name: Optional[str] = None
    # Read secrets from a file in yaml format
    secrets_file: Optional[Path] = None

    # A list of Availability Zones (AZs) where DB instances in the DB cluster can be created.
    availability_zones: Optional[List[str]] = None
    # The number of days for which automated backups are retained.
    # Default: 1
    # Constraints: Must be a value from 1 to 35
    backup_retention_period: Optional[int] = None
    # A value that indicates that the DB cluster should be associated with the specified CharacterSet.
    character_set_name: Optional[str] = None
    # The DB cluster identifier. This parameter is stored as a lowercase string.
    # If None, use the name as the db_cluster_identifier
    # Constraints:
    #   Must contain from 1 to 63 letters, numbers, or hyphens.
    #   First character must be a letter.
    #   Can't end with a hyphen or contain two consecutive hyphens.
    db_cluster_identifier: Optional[str] = None
    # The name of the DB cluster parameter group to associate with this DB cluster.
    # If you do not specify a value, then the default DB cluster parameter group for the specified DB engine and version is used.
    # Constraints: If supplied, must match the name of an existing DB cluster parameter group.
    db_cluster_parameter_group_name: Optional[str] = None

    ## Networking
    # A list of EC2 VPC security groups to associate with this DB cluster.
    vpc_security_group_ids: Optional[List[str]] = None
    # If vpc_security_group_ids is None,
    # Read the security_group_id from vpc_stack
    vpc_stack: Optional[CloudFormationStack] = None

    # A DB subnet group to associate with this DB cluster.
    # This setting is required to create a Multi-AZ DB cluster.
    db_subnet_group_name: Optional[str] = None
    # If db_subnet_group_name is None,
    # Read the db_subnet_group_name from db_subnet_group
    db_subnet_group: Optional[DbSubnetGroup] = None

    # A value that indicates that the DB cluster should be associated with the specified option group.
    option_group_name: Optional[str] = None
    # The daily time range during which automated backups are created if automated backups are enabled
    # using the BackupRetentionPeriod parameter.
    # The default is a 30-minute window selected at random from an 8-hour block of time for each
    # Amazon Web Services Region.
    # Constraints:
    #   Must be in the format hh24:mi-hh24:mi .
    #   Must be in Universal Coordinated Time (UTC).
    #   Must not conflict with the preferred maintenance window.
    #   Must be at least 30 minutes.
    preferred_backup_window: Optional[str] = None
    # The weekly time range during which system maintenance can occur, in Universal Coordinated Time (UTC).
    # Format: ddd:hh24:mi-ddd:hh24:mi
    # The default is a 30-minute window selected at random from an 8-hour block of time for each
    # Amazon Web Services Region, occurring on a random day of the week.
    # Valid Days: Mon, Tue, Wed, Thu, Fri, Sat, Sun.
    # Constraints: Minimum 30-minute window.
    preferred_maintenance_window: Optional[str] = None
    # The Amazon Resource Name (ARN) of the source DB instance or DB cluster
    # if this DB cluster is created as a read replica.
    replication_source_identifier: Optional[str] = None
    # Tags to assign to the DB cluster.
    tags: Optional[List[Dict[str, str]]] = None
    # A value that indicates whether the DB cluster is encrypted.
    storage_encrypted: Optional[bool] = None
    # The Amazon Web Services KMS key identifier for an encrypted DB cluster.
    kms_key_id: Optional[str] = None
    # A value that indicates whether to enable mapping of Amazon Web Services Identity and Access Management (IAM)
    # accounts to database accounts. By default, mapping isn't enabled.
    enable_iamdatabase_authentication: Optional[bool] = None
    # The target backtrack window, in seconds. To disable backtracking, set this value to 0.
    # Default: 0
    backtrack_window: Optional[int] = None
    # The list of log types that need to be enabled for exporting to CloudWatch Logs.
    # The values in the list depend on the DB engine being used.
    # RDS for MySQL: Possible values are error , general , and slowquery .
    # RDS for PostgreSQL: Possible values are postgresql and upgrade .
    # Aurora MySQL: Possible values are audit , error , general , and slowquery .
    # Aurora PostgreSQL: Possible value is postgresql .
    enable_cloudwatch_logs_exports: Optional[List[str]] = None
    # The DB engine mode of the DB cluster,
    # either provisioned , serverless , parallelquery , global , or multimaster .
    engine_mode: Optional[
        Literal["provisioned", "serverless", "parallelquery", "global", "multimaster"]
    ] = None
    # For DB clusters in serverless DB engine mode, the scaling properties of the DB cluster.
    scaling_configuration: Optional[Dict[str, Any]] = None
    # A value that indicates whether the DB cluster has deletion protection enabled.
    # The database can't be deleted when deletion protection is enabled. By default, deletion protection isn't enabled.
    deletion_protection: Optional[bool] = None
    # The global cluster ID of an Aurora cluster that becomes the primary cluster in the new global database cluster.
    global_cluster_identifier: Optional[str] = None
    # A value that indicates whether to enable the HTTP endpoint for an Aurora Serverless v1 DB cluster.
    # By default, the HTTP endpoint is disabled.
    # When enabled, the HTTP endpoint provides a connectionless web service API for running SQL queries on the
    # Aurora Serverless v1 DB cluster. You can also query your database from inside the RDS console with the query editor.
    enable_http_endpoint: Optional[bool] = None
    # A value that indicates whether to copy all tags from the DB cluster to snapshots of the DB cluster.
    # The default is not to copy them.
    copy_tags_to_snapshot: Optional[bool] = None
    # The Active Directory directory ID to create the DB cluster in.
    # For Amazon Aurora DB clusters, Amazon RDS can use Kerberos authentication to authenticate users that connect to
    # the DB cluster.
    domain: Optional[str] = None
    # Specify the name of the IAM role to be used when making API calls to the Directory Service.
    domain_iamrole_name: Optional[str] = None
    enable_global_write_forwarding: Optional[bool] = None
    # The amount of storage in gibibytes (GiB) to allocate to each DB instance in the Multi-AZ DB cluster.
    allocated_storage: Optional[int] = None
    # Specifies the storage type to be associated with the DB cluster.
    # This setting is required to create a Multi-AZ DB cluster.
    # Valid values: io1
    # When specified, a value for the Iops parameter is required.
    # Default: io1
    storage_type: Optional[str] = None
    # The amount of Provisioned IOPS (input/output operations per second) to be initially allocated for each DB
    # instance in the Multi-AZ DB cluster.
    iops: Optional[int] = None
    # A value that indicates whether the DB cluster is publicly accessible.
    # When the DB cluster is publicly accessible, its Domain Name System (DNS) endpoint resolves to the private IP
    # address from within the DB cluster's virtual private cloud (VPC). It resolves to the public IP address from
    # outside of the DB cluster's VPC. Access to the DB cluster is ultimately controlled by the security group it uses.
    # That public access isn't permitted if the security group assigned to the DB cluster doesn't permit it.
    # When the DB cluster isn't publicly accessible, it is an internal DB cluster with a DNS name
    # that resolves to a private IP address.
    publicly_accessible: Optional[bool] = None
    # A value that indicates whether minor engine upgrades are applied automatically to the DB cluster during the
    # maintenance window. By default, minor engine upgrades are applied automatically.
    auto_minor_version_upgrade: Optional[bool] = None
    # The interval, in seconds, between points when Enhanced Monitoring metrics are collected for the DB cluster.
    # To turn off collecting Enhanced Monitoring metrics, specify 0. The default is 0.
    # If MonitoringRoleArn is specified, also set MonitoringInterval to a value other than 0.
    # Valid Values: 0, 1, 5, 10, 15, 30, 60
    monitoring_interval: Optional[int] = None
    # The Amazon Resource Name (ARN) for the IAM role that permits RDS to send
    # Enhanced Monitoring metrics to Amazon CloudWatch Logs.
    monitoring_role_arn: Optional[str] = None
    enable_performance_insights: Optional[bool] = None
    performance_insights_kmskey_id: Optional[str] = None
    performance_insights_retention_period: Optional[int] = None
    # The ID of the region that contains the source for the db cluster.
    source_region: Optional[str] = None

    def get_db_cluster_identifier(self):
        return self.db_cluster_identifier or self.name

    def get_secret_data_from_file(self) -> Optional[Dict[str, str]]:
        import yaml

        secrets_file_path = self.secrets_file
        if (
            secrets_file_path is not None
            and secrets_file_path.exists()
            and secrets_file_path.is_file()
        ):
            # logger.debug(f"Reading {secrets_file_path}")
            secret_data_from_file = yaml.safe_load(secrets_file_path.read_text())
            if secret_data_from_file is not None and isinstance(
                secret_data_from_file, dict
            ):
                return secret_data_from_file
            else:
                print_error(f"Invalid secrets_file: {secrets_file_path}")
        return None

    def get_master_username(self) -> Optional[str]:
        master_username_var: Optional[str] = self.master_username if self else None
        if master_username_var is None and self.secrets_file is not None:
            # read from secrets_file
            # logger.debug(f"Reading POSTGRES_USER from secrets_file")
            secret_data_from_file = self.get_secret_data_from_file()
            if secret_data_from_file is not None:
                master_username_var = secret_data_from_file.get(
                    "MASTER_USERNAME", master_username_var
                )
        return master_username_var

    def get_master_user_password(self) -> Optional[str]:
        master_user_password_var: Optional[str] = (
            self.master_user_password if self else None
        )
        if master_user_password_var is None and self.secrets_file is not None:
            # read from secrets_file
            # logger.debug(f"Reading POSTGRES_USER from secrets_file")
            secret_data_from_file = self.get_secret_data_from_file()
            if secret_data_from_file is not None:
                master_user_password_var = secret_data_from_file.get(
                    "MASTER_USER_PASSWORD", master_user_password_var
                )
        return master_user_password_var

    def get_database_name(self) -> Optional[str]:
        database_name_var: Optional[str] = self.database_name if self else None
        if database_name_var is None and self.secrets_file is not None:
            # read from secrets_file
            # logger.debug(f"Reading POSTGRES_USER from secrets_file")
            secret_data_from_file = self.get_secret_data_from_file()
            if secret_data_from_file is not None:
                database_name_var = secret_data_from_file.get(
                    "DATABASE_NAME", database_name_var
                )
        return database_name_var

    def _create(self, aws_client: AwsApiClient) -> bool:
        """Creates the DbCluster

        Args:
            aws_client: The AwsApiClient for the current cluster
        """

        print_info(f"Creating {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            # create a dict of args which are not null, otherwise aws type validation fails
            not_null_args = {}

            not_null_args["DBClusterIdentifier"] = self.get_db_cluster_identifier()
            if self.engine_version:
                not_null_args["EngineVersion"] = self.engine_version
            if self.port:
                not_null_args["Port"] = self.port
            if self.master_username:
                not_null_args["MasterUsername"] = self.master_username
            if self.master_user_password:
                not_null_args["MasterUserPassword"] = self.master_user_password
            if self.availability_zones:
                not_null_args["AvailabilityZones"] = self.availability_zones
            if self.backup_retention_period:
                not_null_args["BackupRetentionPeriod"] = self.backup_retention_period
            if self.character_set_name:
                not_null_args["CharacterSetName"] = self.character_set_name
            if self.database_name:
                not_null_args["DatabaseName"] = self.database_name
            if self.db_cluster_parameter_group_name:
                not_null_args[
                    "DBClusterParameterGroupName"
                ] = self.db_cluster_parameter_group_name
            if self.option_group_name:
                not_null_args["OptionGroupName"] = self.option_group_name
            if self.preferred_backup_window:
                not_null_args["PreferredBackupWindow"] = self.preferred_backup_window
            if self.preferred_maintenance_window:
                not_null_args[
                    "PreferredMaintenanceWindow"
                ] = self.preferred_maintenance_window
            if self.replication_source_identifier:
                not_null_args[
                    "ReplicationSourceIdentifier"
                ] = self.replication_source_identifier
            if self.tags:
                not_null_args["Tags"] = self.tags
            if self.storage_encrypted:
                not_null_args["StorageEncrypted"] = self.storage_encrypted
            if self.kms_key_id:
                not_null_args["KmsKeyId"] = self.kms_key_id
            if self.enable_iamdatabase_authentication:
                not_null_args[
                    "EnableIamdatabaseAuthentication"
                ] = self.enable_iamdatabase_authentication
            if self.backtrack_window:
                not_null_args["BacktrackWindow"] = self.backtrack_window
            if self.enable_cloudwatch_logs_exports:
                not_null_args[
                    "EnableCloudwatchLogsExports"
                ] = self.enable_cloudwatch_logs_exports
            if self.engine_mode:
                not_null_args["EngineMode"] = self.engine_mode
            if self.scaling_configuration:
                not_null_args["ScalingConfiguration"] = self.scaling_configuration
            if self.deletion_protection:
                not_null_args["DeletionProtection"] = self.deletion_protection
            if self.global_cluster_identifier:
                not_null_args[
                    "GlobalClusterIdentifier"
                ] = self.global_cluster_identifier
            if self.enable_http_endpoint:
                not_null_args["EnableHttpEndpoint"] = self.enable_http_endpoint
            if self.copy_tags_to_snapshot:
                not_null_args["CopyTagsToSnapshot"] = self.copy_tags_to_snapshot
            if self.domain:
                not_null_args["Domain"] = self.domain
            if self.domain_iamrole_name:
                not_null_args["DomainIAMRoleName"] = self.domain_iamrole_name
            if self.enable_global_write_forwarding:
                not_null_args[
                    "EnableGlobalWriteForwarding"
                ] = self.enable_global_write_forwarding
            if self.db_instance_class:
                not_null_args["DBClusterInstanceClass"] = self.db_instance_class
            if self.allocated_storage:
                not_null_args["AllocatedStorage"] = self.allocated_storage
            if self.storage_type:
                not_null_args["StorageType"] = self.storage_type
            if self.iops:
                not_null_args["Iops"] = self.iops
            if self.publicly_accessible:
                not_null_args["PubliclyAccessible"] = self.publicly_accessible
            if self.auto_minor_version_upgrade:
                not_null_args[
                    "AutoMinorVersionUpgrade"
                ] = self.auto_minor_version_upgrade
            if self.monitoring_interval:
                not_null_args["MonitoringInterval"] = self.monitoring_interval
            if self.monitoring_role_arn:
                not_null_args["MonitoringRoleArn"] = self.monitoring_role_arn
            if self.enable_performance_insights:
                not_null_args[
                    "EnablePerformanceInsights"
                ] = self.enable_performance_insights
            if self.performance_insights_kmskey_id:
                not_null_args[
                    "PerformanceInsightsKMSKeyId"
                ] = self.performance_insights_kmskey_id
            if self.performance_insights_retention_period:
                not_null_args[
                    "PerformanceInsightsRetentionPeriod"
                ] = self.performance_insights_retention_period
            if self.source_region:
                not_null_args["SourceRegion"] = self.source_region

            vpc_security_group_ids = self.vpc_security_group_ids
            if vpc_security_group_ids is None and self.vpc_stack is not None:
                vpc_stack_sg = self.vpc_stack.get_security_group(aws_client=aws_client)
                if vpc_stack_sg is not None:
                    logger.debug(f"Using SecurityGroup: {vpc_stack_sg}")
                    vpc_security_group_ids = [vpc_stack_sg]
            if vpc_security_group_ids is not None:
                not_null_args["VpcSecurityGroupIds"] = vpc_security_group_ids

            db_subnet_group_name = self.db_subnet_group_name
            if db_subnet_group_name is None and self.db_subnet_group is not None:
                db_subnet_group_name = self.db_subnet_group.name
                logger.debug(f"Using DbSubnetGroup: {db_subnet_group_name}")
            if db_subnet_group_name is not None:
                not_null_args["DBSubnetGroupName"] = db_subnet_group_name

            # Create DbCluster
            service_client = self.get_service_client(aws_client)
            create_response = service_client.create_db_cluster(
                Engine=self.engine,
                **not_null_args,
            )
            logger.debug(f"create_response type: {type(create_response)}")
            logger.debug(f"create_response: {create_response}")

            self.active_resource = create_response.get("DBCluster", None)
            if self.active_resource is not None:
                print_info(
                    f"{self.get_resource_type()}: {self.get_resource_name()} created"
                )
                logger.debug(f"DbCluster: {self.active_resource}")
                return True
        except Exception as e:
            print_error(f"{self.get_resource_type()} could not be created.")
            print_error("Please try again.")
            print_error(e)
        return False

    def _read(self, aws_client: AwsApiClient) -> Optional[Any]:
        """Returns the DbCluster

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            service_client = self.get_service_client(aws_client)
            db_cluster_identifier = self.get_db_cluster_identifier()
            describe_response = service_client.describe_db_clusters(
                DBClusterIdentifier=db_cluster_identifier
            )
            logger.debug(f"describe_response type: {type(describe_response)}")
            logger.debug(f"describe_response: {describe_response}")

            db_clusters_list = describe_response.get("DBClusters", None)
            if db_clusters_list is not None and isinstance(db_clusters_list, list):
                for _db_cluster in db_clusters_list:
                    _cluster_identifier = _db_cluster.get("DBClusterIdentifier", None)
                    if _cluster_identifier == db_cluster_identifier:
                        self.active_resource = _db_cluster
                        break

            if self.active_resource is None:
                logger.debug(f"No {self.get_resource_type()} found")
                return None

            logger.debug(f"DbCluster: {self.active_resource}")
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
            pass
        except Exception as e:
            print_error(f"Error reading {self.get_resource_type()}.")
            print_error("Please try again.")
            print_error(e)
        return self.active_resource

    def _delete(self, aws_client: AwsApiClient) -> bool:
        """Deletes the DbCluster

        Args:
            aws_client: The AwsApiClient for the current cluster
        """

        print_info(f"Deleting {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            service_client = self.get_service_client(aws_client)
            self.active_resource = None

            db_cluster_identifier = self.get_db_cluster_identifier()
            delete_response = service_client.delete_db_cluster(
                DBClusterIdentifier=db_cluster_identifier
            )
            logger.debug(f"delete_response: {delete_response}")
            print_info(
                f"{self.get_resource_type()}: {self.get_resource_name()} deleted"
            )
            return True
        except Exception as e:
            print_error(f"{self.get_resource_type()} could not be deleted.")
            print_error("Please try again or delete resources manually.")
            print_error(e)
        return False



from typing import List, Optional, Type, Tuple, Set

from typing_extensions import Literal

from phidata.infra.aws.resource.base import AwsResource
from phidata.infra.aws.resource.group import AwsResourceGroup
from phidata.infra.aws.resource.types import (
    AwsResourceInstallOrder,
    AwsResourceType,
)
from phidata.utils.log import logger


def get_install_weight_for_aws_resource(aws_resource_type: AwsResourceType) -> int:
    """Function which returns the install weight for an AwsResource"""

    resource_type_class_name = aws_resource_type.__class__.__name__
    if resource_type_class_name in AwsResourceInstallOrder.keys():
        install_weight = AwsResourceInstallOrder[resource_type_class_name]
        # logger.debug(
        #     "Resource type: {} | Install weight: {}".format(
        #         resource_type_class_name,
        #         install_weight,
        #     )
        # )
        return install_weight
    return 1000


def filter_aws_resources_group(
    aws_resource_group: AwsResourceGroup,
    name_filter: Optional[str] = None,
    type_filter: Optional[str] = None,
) -> Optional[List[AwsResourceType]]:
    """Filters the AwsResourceGroup using name_filter and type_filter.
    This function also flattens any List[AwsResourceType] attributes.
    Eg: it will flatten s3_buckets

    Args:
        aws_resource_group:
        name_filter:
        type_filter:

    Returns:
        Optional[List[AwsResourceType]]: List of filtered and flattened Aws Resources
    """

    # List of all resources in this AwsResourceGroup
    all_resources: List[AwsResourceType] = []
    # List of filtered resources
    filtered_resources: List[AwsResourceType] = []
    logger.debug(f"Filtering AwsResourceGroup {aws_resource_group.name}")

    # Populate all_resources list
    for resource, resource_data in aws_resource_group.__dict__.items():
        # logger.debug("resource: {}".format(resource))

        ## Check if the resource is a single AwsResourceType or a List[AwsResourceType]

        # If the resource is a List[AwsResourceType], flatten the resources, verify each element
        # is isinstance of AwsResource and add to the all_resources list
        if isinstance(resource_data, list):
            for _r in resource_data:
                if isinstance(_r, AwsResource):
                    all_resources.append(_r)
        # If the resource is a single resource, verify that the resource isinstance of
        # AwsResource and add it to the all_resources list
        elif isinstance(resource_data, AwsResource):
            all_resources.append(resource_data)

    # Apply filters
    if name_filter is not None or type_filter is not None:
        for resource in all_resources:
            resource_name = resource.get_resource_name()
            resource_type = resource.get_resource_type()
            # skip resources that dont match the name_filter
            if name_filter is not None:
                if resource_name is None:
                    logger.debug(f"Skipping unnamed resource, type: {resource_type}")
                    continue
                if name_filter.lower() not in resource_name.lower():
                    logger.debug(f"Skipping {resource_type} | {resource_name}")
                    continue
            # skip resources that dont match the type_filter
            if type_filter is not None:
                resource_type = resource.get_resource_type()
                if resource_type is None:
                    logger.debug(f"Skipping untyped resource, name: {resource_name}")
                    continue
                if type_filter.lower() not in resource_type.lower():
                    logger.debug(f"Skipping {resource_type} | {resource_name}")
                    continue
            filtered_resources.append(resource)
    else:
        filtered_resources = all_resources

    return filtered_resources


# def dedup_resource_types(
#     aws_resources: Optional[List[AwsResourceType]] = None,
# ) -> Optional[Set[Type[AwsResource]]]:
#     """Takes a list of AwsResources and returns a Set of AwsResource classes.
#     Each AwsResource classes is represented by the Type[resources.AwsResource] type.
#     From python docs:
#         A variable annotated with Type[AwsResource] may accept values that are classes.
#         Acceptable classes are the AwsResource class + subclasses.
#     """
#     if aws_resources:
#         active_resource_types: Set[Type[AwsResource]] = set()
#         for resource in aws_resources:
#             active_resource_types.add(resource.__class__)
#             # logger.debug(f"Gathering: {resource.get_resource_name()}")
#             # logger.debug(f"Resource Type: {resource_type}")
#         logger.debug("Active Resource Types: {}".format(active_resource_types))
#         return active_resource_types
#     return None


def filter_and_flatten_aws_resource_group(
    aws_resource_group: AwsResourceGroup,
    name_filter: Optional[str] = None,
    type_filter: Optional[str] = None,
    sort_order: Literal["create", "delete"] = "create",
) -> Optional[List[AwsResourceType]]:
    """This function flattens the aws_resource_group and returns a filtered list of
    AwsResources sorted in the order requested. create == install order, delete == reverse

    Args:
        aws_resource_group: AwsResourceGroup
        name_filter: Filter AwsResources by name
        type_filter: Filter AwsResources by type
        sort_order: create or delete

    Returns:
        List[AwsResourceType]: List of filtered Aws Resources
    """

    filtered_aws_resources: List[AwsResourceType] = []
    if aws_resource_group is not None and aws_resource_group.enabled:
        # logger.debug(f"Flattening AwsResourceGroup {aws_resource_group.name}")
        filtered_aws_resources = filter_aws_resources_group(
            aws_resource_group, name_filter, type_filter
        )

    # Sort the resources in install order
    if sort_order == "create":
        filtered_aws_resources.sort(
            key=lambda x: get_install_weight_for_aws_resource(x)
        )
    elif sort_order == "delete":
        filtered_aws_resources.sort(
            key=lambda x: get_install_weight_for_aws_resource(x), reverse=True
        )

    # logger.debug("filtered_aws_resources: {}".format(filtered_aws_resources))
    return filtered_aws_resources



from pathlib import Path
from typing import Optional, Literal, Union, List

from phidata.asset.file import File, FileType
from phidata.asset.table.sql import SqlTable
from phidata.utils.log import logger
from phidata.workflow.py.python_workflow import (
    PythonWorkflow,
    PythonWorkflowArgs,
    EngineType,
)


class UploadFileToSqlArgs(PythonWorkflowArgs):
    file: File
    sql_table: SqlTable
    # How to behave if the sql_table already exists.
    # fail: Raise a ValueError.
    # replace: Drop the table before inserting new values.
    # append: Insert new values to the existing table.
    if_exists: Optional[Literal["fail", "replace", "append"]] = None
    # Write DataFrame index as a column.
    # Uses index_label as the column name in the table.
    index: Optional[bool] = None
    # Column label for index column(s).
    # If None is given (default) and index is True, then the index names are used.
    # A sequence should be given if the DataFrame uses MultiIndex.
    index_label: Optional[Union[str, List[str]]] = None
    # Specify the number of rows in each batch to be written at a time.
    # By default, all rows will be written at once.
    chunksize: Optional[int] = None
    create_db_engine_from_conn_id: bool = True


class UploadFileToSql(PythonWorkflow):
    def __init__(
        self,
        file: File,
        sql_table: SqlTable,
        if_exists: Optional[Literal["fail", "replace", "append"]] = None,
        index: Optional[bool] = None,
        index_label: Optional[Union[str, List[str]]] = None,
        chunksize: Optional[int] = None,
        engine: EngineType = EngineType.PANDAS,
        name: str = "load_file_to_sql",
        task_id: Optional[str] = None,
        dag_id: Optional[str] = None,
        version: Optional[str] = None,
        enabled: bool = True,
    ):
        super().__init__()
        try:
            self.args: UploadFileToSqlArgs = UploadFileToSqlArgs(
                file=file,
                sql_table=sql_table,
                if_exists=if_exists,
                index=index,
                index_label=index_label,
                chunksize=chunksize,
                engine=engine,
                name=name,
                task_id=task_id,
                dag_id=dag_id,
                version=version,
                enabled=enabled,
                entrypoint=load_file_to_sql_table,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    @property
    def file(self) -> Optional[File]:
        return self.args.file

    @file.setter
    def file(self, file: File) -> None:
        if file is not None:
            self.args.file = file

    @property
    def sql_table(self) -> Optional[SqlTable]:
        return self.args.sql_table

    @sql_table.setter
    def sql_table(self, sql_table: SqlTable) -> None:
        if sql_table is not None:
            self.args.sql_table = sql_table


def load_file_to_sql_table_pandas(args: UploadFileToSqlArgs) -> bool:

    import pandas as pd

    file_to_load: Optional[File] = args.file
    if file_to_load is None:
        logger.error("File not available")
        return False
    file_path: Optional[Path] = file_to_load.file_path
    if file_path is None:
        logger.error("FilePath not available")
        return False
    file_type: Optional[FileType] = file_to_load.file_type
    if file_type is None:
        logger.error("FileType not available")
        return False
    # logger.debug("File: {}".format(file_to_load.args))

    sql_table: Optional[SqlTable] = args.sql_table
    if sql_table is None:
        logger.error("SqlTable not available")
        return False
    # logger.debug("SqlTable: {}".format(sql_table.args))

    logger.info("Reading: {}".format(file_path))
    df: Optional[pd.DataFrame] = None
    if file_type == FileType.CSV:
        df = pd.read_csv(file_path)
    elif file_type == FileType.JSON:
        df = pd.read_json(file_path)
    elif file_type == FileType.TSV:
        df = pd.read_csv(file_path, sep="\t")

    if df is not None:
        # logger.info()("DataFrame:\n{}".format(df.head()))
        logger.info("Writing to table: {}".format(sql_table.name))
        upload_success = sql_table.write_pandas_df(
            df=df,
            if_exists=args.if_exists,
            index=args.index,
            index_label=args.index_label,
            chunksize=args.chunksize,
        )
        return upload_success
    else:
        logger.info("Could not read file into DataFrame")
        return False


def load_file_to_sql_table(**kwargs) -> bool:
    args: UploadFileToSqlArgs = UploadFileToSqlArgs(**kwargs)
    # logger.debug("LoadFileToSqlTableArgs: {}".format(args))

    if args.engine in (EngineType.PANDAS, EngineType.DEFAULT):
        return load_file_to_sql_table_pandas(args)
    else:
        logger.error(f"EngineType: {args.engine} not yet supported")
        return False



from pathlib import Path
from typing import Optional, List, Dict

from phidata.infra.aws.args import AwsArgs
from phidata.infra.aws.enums import AwsManagerStatus
from phidata.infra.aws.exceptions import AwsArgsException
from phidata.infra.aws.resource.types import AwsResourceType
from phidata.infra.aws.worker import AwsWorker
from phidata.infra.k8s.args import K8sArgs
from phidata.infra.k8s.exceptions import K8sWorkerException
from phidata.infra.k8s.worker import K8sWorker
from phidata.utils.cli_console import print_error, print_info
from phidata.utils.log import logger


class AwsManager:
    def __init__(self, aws_args: AwsArgs):
        if aws_args is None or not isinstance(aws_args, AwsArgs):
            raise AwsArgsException("Invalid AwsArgs")
        if aws_args.workspace_root_path is None or not isinstance(
            aws_args.workspace_root_path, Path
        ):
            raise AwsArgsException("workspace_root_path invalid")
        if aws_args.workspace_config_file_path is None or not isinstance(
            aws_args.workspace_config_file_path, Path
        ):
            raise AwsArgsException("workspace_config_file_path invalid")

        self.aws_args: AwsArgs = aws_args
        self.aws_worker: Optional[AwsWorker] = AwsWorker(self.aws_args)
        self.aws_status: AwsManagerStatus = AwsManagerStatus.PRE_INIT
        logger.debug("**-+-** AwsManager created")

    def get_status(self, refresh: bool = False) -> AwsManagerStatus:
        # logger.debug("Getting AwsManagerStatus")
        if refresh:
            self.aws_status = AwsManagerStatus.PRE_INIT
            logger.debug(f"AwsManagerStatus: {self.aws_status.value}")

        if self.aws_status == AwsManagerStatus.PRE_INIT:
            if self.aws_worker is not None and self.aws_worker.is_client_initialized():
                self.aws_status = AwsManagerStatus.WORKER_READY

        if self.aws_status == AwsManagerStatus.WORKER_READY:
            if (
                self.aws_worker is not None
                and self.aws_worker.are_resources_initialized()
            ):
                self.aws_status = AwsManagerStatus.RESOURCES_INIT

        logger.debug(f"AwsManagerStatus: {self.aws_status.value}")
        return self.aws_status

    ######################################################
    ## Create Resources
    ######################################################

    def create_resources_dry_run(
        self, name_filter: Optional[str] = None, type_filter: Optional[str] = None
    ) -> None:

        status = self.get_status()
        if not status.can_create_resources():
            print_error("Cannot create AwsResources")
            return
        if self.aws_worker is None:
            print_error("AWSWorker not available")
            return

        # logger.debug("Creating resources dry run")
        self.aws_worker.create_resources_dry_run(
            name_filter=name_filter, type_filter=type_filter
        )

    def create_resources(
        self, name_filter: Optional[str] = None, type_filter: Optional[str] = None
    ) -> bool:

        status = self.get_status()
        if not status.can_create_resources():
            logger.debug("Cannot create AwsResources")
            return False
        if self.aws_worker is None:
            logger.debug("AWSWorker not available")
            return False

        return self.aws_worker.create_resources(
            name_filter=name_filter, type_filter=type_filter
        )

    def validate_resources_are_created(
        self, name_filter: Optional[str] = None, type_filter: Optional[str] = None
    ) -> bool:

        logger.debug("Validating resources are created...")
        return True

    ######################################################
    ## Read Resources
    ######################################################

    def read_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
    ) -> Dict[str, List[AwsResourceType]]:

        status = self.get_status()
        if not status.can_get_resources():
            logger.debug("Cannot get AwsResources")
            return None
        if self.aws_worker is None:
            logger.debug("AWSWorker not available")
            return None

        return self.aws_worker.read_resources(name_filter, type_filter)

    ######################################################
    ## Delete Resources
    ######################################################

    def delete_resources_dry_run(
        self, name_filter: Optional[str] = None, type_filter: Optional[str] = None
    ) -> None:

        status = self.get_status()
        if not status.can_delete_resources():
            logger.debug("Cannot delete AwsResources")
            return
        if self.aws_worker is None:
            logger.debug("AWSWorker not available")
            return

        # logger.debug("Deleting resources dry run")
        self.aws_worker.delete_resources_dry_run(
            name_filter=name_filter, type_filter=type_filter
        )

    def delete_resources(
        self, name_filter: Optional[str] = None, type_filter: Optional[str] = None
    ) -> bool:

        status = self.get_status()
        if not status.can_delete_resources():
            logger.debug("Cannot delete AwsResources")
            return False
        if self.aws_worker is None:
            logger.debug("AWSWorker not available")
            return False

        return self.aws_worker.delete_resources(
            name_filter=name_filter, type_filter=type_filter
        )

    def validate_resources_are_deleted(
        self, name_filter: Optional[str] = None, type_filter: Optional[str] = None
    ) -> bool:

        logger.debug("Validating resources are deleted...")
        return True

    ######################################################
    ## Patch Resources
    ######################################################

    def patch_resources_dry_run(
        self, name_filter: Optional[str] = None, type_filter: Optional[str] = None
    ) -> None:

        status = self.get_status()
        if not status.can_create_resources():
            logger.debug("Cannot patch AwsResources")
            return
        if self.aws_worker is None:
            logger.debug("AWSWorker not available")
            return

        # logger.debug("Deleting resources dry run")
        self.aws_worker.patch_resources_dry_run(
            name_filter=name_filter, type_filter=type_filter
        )

    def patch_resources(
        self, name_filter: Optional[str] = None, type_filter: Optional[str] = None
    ) -> bool:

        status = self.get_status()
        if not status.can_create_resources():
            logger.debug("Cannot patch AwsResources")
            return False
        if self.aws_worker is None:
            logger.debug("AWSWorker not available")
            return False

        return self.aws_worker.patch_resources(
            name_filter=name_filter, type_filter=type_filter
        )

    def validate_resources_are_patched(
        self, name_filter: Optional[str] = None, type_filter: Optional[str] = None
    ) -> bool:

        logger.debug("Validating resources are patched...")
        return True



from typing import Optional

from phidata.asset import DataAsset, DataAssetArgs


class AthenaTableArgs(DataAssetArgs):
    name: str
    database: Optional[str] = None
    catalog: Optional[str] = None
    work_group: Optional[str] = None


class AthenaTable(DataAsset):
    def __init__(
        self,
        name: str,
        database: Optional[str] = None,
        catalog: Optional[str] = None,
        work_group: Optional[str] = None,
        version: Optional[str] = None,
        enabled: bool = True,
    ) -> None:

        super().__init__()
        try:
            self.args: AthenaTableArgs = AthenaTableArgs(
                name=name,
                database=database,
                catalog=catalog,
                work_group=work_group,
                version=version,
                enabled=enabled,
            )
        except Exception as e:
            raise

    @property
    def name(self) -> str:
        return self.args.name

    @property
    def database(self) -> Optional[str]:
        return self.args.database

    @property
    def catalog(self) -> Optional[str]:
        return self.args.catalog

    @property
    def work_group(self) -> Optional[str]:
        return self.args.work_group



from collections import OrderedDict
from pathlib import Path
from typing import Optional, Dict, Any, List, Union
from typing_extensions import Literal

from pydantic import BaseModel

from phidata.app.db import DbApp
from phidata.app import PhidataApp, PhidataAppArgs
from phidata.infra.docker.resource.network import DockerNetwork
from phidata.infra.docker.resource.container import DockerContainer
from phidata.infra.docker.resource.group import (
    DockerResourceGroup,
    DockerBuildContext,
)
from phidata.infra.k8s.create.apps.v1.deployment import CreateDeployment
from phidata.infra.k8s.create.core.v1.config_map import CreateConfigMap
from phidata.infra.k8s.create.core.v1.container import CreateContainer
from phidata.infra.k8s.create.core.v1.volume import (
    CreateVolume,
    EmptyDirVolumeSource,
    HostPathVolumeSource,
    VolumeType,
)
from phidata.infra.k8s.create.common.port import CreatePort
from phidata.infra.k8s.create.group import CreateK8sResourceGroup
from phidata.infra.k8s.resource.group import (
    K8sResourceGroup,
    K8sBuildContext,
)
from phidata.utils.common import (
    get_image_str,
    get_default_container_name,
    get_default_configmap_name,
    get_default_secret_name,
    get_default_service_name,
    get_default_volume_name,
    get_default_deploy_name,
    get_default_pod_name,
)
from phidata.utils.cli_console import print_error, print_info
from phidata.utils.log import logger

default_devbox_name: str = "devbox"


class DevboxDevModeArgs(BaseModel):
    install_phidata_dev: bool = True
    phidata_volume_name: str = "devbox-phidata-volume"
    phidata_dir_path: Path = Path.home().joinpath("p", "philab", "phidata")
    phidata_dir_container_path: str = "/phidata"


class DevboxArgs(PhidataAppArgs):
    name: str = default_devbox_name
    version: str = "1"
    enabled: bool = True

    # Image args
    image_name: str = "phidata/devbox"
    image_tag: str = "1"
    entrypoint: Optional[Union[str, List]] = None
    command: Optional[Union[str, List]] = None

    # Mount the workspace directory on the container
    mount_workspace: bool = True
    workspace_volume_name: str = "devbox-workspace-volume"
    # Path to mount the workspace volume under
    # This is the parent directory for the workspace on the container
    # i.e. the ws is mounted as a subdir in this dir
    # eg: if ws name is: idata, workspace path would be: /usr/local/devbox/idata
    workspace_parent_container_path: str = "/usr/local/devbox"

    # Install python dependencies using a requirements.txt file
    install_requirements: bool = True
    # Path to the requirements.txt file relative to the workspace root
    requirements_file_path: str = "requirements.txt"

    # Mount aws config on the container
    mount_aws_config: bool = False
    # Aws config dir on the host
    aws_config_path: Path = Path.home().resolve().joinpath(".aws")
    # Aws config dir on the container
    aws_config_container_path: str = "/root/.aws"

    # Configure airflow
    # If init_airflow = True, this devbox initializes airflow
    init_airflow: bool = True
    # If use_products_as_airflow_dags = True
    # set the AIRFLOW__CORE__DAGS_FOLDER to the products_dir
    use_products_as_airflow_dags: bool = True
    # If use_products_as_airflow_dags = False
    # set the AIRFLOW__CORE__DAGS_FOLDER to the airflow_dags_path
    # airflow_dags_path is the directory in the container containing the airflow dags
    airflow_dags_path: Optional[str] = None
    # Creates an airflow user with username: test, pass: test
    create_airflow_test_user: bool = False
    airflow_executor: Literal[
        "DebugExecutor",
        "LocalExecutor",
        "SequentialExecutor",
        "CeleryExecutor",
        "CeleryKubernetesExecutor",
        "DaskExecutor",
        "KubernetesExecutor",
    ] = "SequentialExecutor"

    # Only on docker: mount airflow home from container to host machine
    # Useful when debugging the airflow conf
    mount_airflow_home: bool = False
    # Path to the dir on host machine relative to the workspace root
    airflow_home_dir: str = "airflow"
    # Path to airflow home
    airflow_home_container_path: str = "/usr/local/airflow"

    # Configure airflow db
    # If True, initialize the airflow_db on this devbox
    # If None, value is derived from init_airflow i.e. initialize the airflow_db if init_airflow = True
    #   Locally, airflow_db uses sqllite
    # If using the devbox with an external Airflow db, set init_airflow_db = False
    init_airflow_db: Optional[bool] = None
    wait_for_airflow_db: bool = False
    # Connect to database using DbApp
    airflow_db_app: Optional[DbApp] = None
    # Connect to database manually
    airflow_db_user: Optional[str] = None
    airflow_db_password: Optional[str] = None
    airflow_db_schema: Optional[str] = None
    airflow_db_host: Optional[str] = None
    airflow_db_port: Optional[str] = None
    airflow_db_driver: str = "postgresql+psycopg2"
    # Airflow db connections in the format { conn_id: conn_url }
    # converted to env var: AIRFLOW_CONN__conn_id = conn_url
    db_connections: Optional[Dict] = None

    # Configure airflow scheduler
    # Init Airflow scheduler as a deamon process
    init_airflow_scheduler: bool = False

    # Configure airflow webserver
    # Init Airflow webserver when the container starts
    init_airflow_webserver: bool = False
    # webserver port on the container
    airflow_webserver_container_port: int = 8080
    # webserver port on the host machine
    airflow_webserver_host_port: int = 8080
    # webserver port name on K8sContainer
    airflow_webserver_port_name: str = "http"

    # Configure the container
    container_name: Optional[str] = None
    # Only used by the DockerContainer
    container_detach: bool = True
    container_auto_remove: bool = True
    container_remove: bool = True
    # Add env variables to container env
    env: Optional[Dict[str, str]] = None
    # Read env variables from a file in yaml format
    env_file: Optional[Path] = None
    # Configure the ConfigMap used for env variables that are not Secret
    config_map_name: Optional[str] = None
    # Configure the Secret used for env variables that are Secret
    secret_name: Optional[str] = None
    # Read secrets from a file in yaml format
    secrets_file: Optional[Path] = None

    # Configure the databox deploy
    deploy_name: Optional[str] = None
    pod_name: Optional[str] = None

    # Other args
    load_examples: bool = False
    print_env_on_load: bool = True
    dev_mode: Optional[DevboxDevModeArgs] = None


class Devbox(PhidataApp):
    def __init__(
        self,
        name: str = default_devbox_name,
        version: str = "1",
        enabled: bool = True,
        # Image args,
        image_name: str = "phidata/devbox",
        image_tag: str = "1",
        entrypoint: Optional[Union[str, List]] = None,
        command: Optional[Union[str, List]] = None,
        # Mount the workspace directory on the container,
        mount_workspace: bool = True,
        workspace_volume_name: str = "devbox-workspace-volume",
        # Path to mount the workspace volume under,
        # This is the parent directory for the workspace on the container,
        # i.e. the ws is mounted as a subdir in this dir,
        # eg: if ws name is: idata, workspace path would be: /usr/local/devbox/idata,
        workspace_parent_container_path: str = "/usr/local/devbox",
        # Install python dependencies using a requirements.txt file,
        install_requirements: bool = True,
        # Path to the requirements.txt file relative to the workspace root,
        requirements_file_path: str = "requirements.txt",
        # Mount aws config on the container,
        mount_aws_config: bool = False,
        # Aws config dir on the host,
        aws_config_path: Path = Path.home().resolve().joinpath(".aws"),
        # Aws config dir on the container,
        aws_config_container_path: str = "/root/.aws",
        # Configure airflow,
        # If init_airflow = True, this devbox initializes airflow,
        init_airflow: bool = True,
        # If use_products_as_airflow_dags = True,
        # set the AIRFLOW__CORE__DAGS_FOLDER to the products_dir,
        use_products_as_airflow_dags: bool = True,
        # If use_products_as_airflow_dags = False,
        # set the AIRFLOW__CORE__DAGS_FOLDER to the airflow_dags_path,
        # airflow_dags_path is the directory in the container containing the airflow dags,
        airflow_dags_path: Optional[str] = None,
        # Creates an airflow user with username: test, pass: test,
        create_airflow_test_user: bool = False,
        # Only on docker: mount airflow home from container to host machine,
        # Useful when debugging the airflow conf,
        mount_airflow_home: bool = False,
        airflow_executor: Literal[
            "DebugExecutor",
            "LocalExecutor",
            "SequentialExecutor",
            "CeleryExecutor",
            "CeleryKubernetesExecutor",
            "DaskExecutor",
            "KubernetesExecutor",
        ] = "SequentialExecutor",
        # Path to the dir on host machine relative to the workspace root,
        airflow_home_dir: str = "airflow",
        # Path to airflow home,
        airflow_home_container_path: str = "/usr/local/airflow",
        # Configure airflow db,
        # If True, initialize the airflow_db on this devbox,
        # If None, value is derived from init_airflow i.e. initialize the airflow_db if init_airflow = True,
        #   Locally, airflow_db uses sqllite,
        # If using the devbox with an external Airflow db, set init_airflow_db = False,
        init_airflow_db: Optional[bool] = None,
        wait_for_airflow_db: bool = False,
        # Connect to database using DbApp,
        airflow_db_app: Optional[DbApp] = None,
        # Connect to database manually,
        airflow_db_user: Optional[str] = None,
        airflow_db_password: Optional[str] = None,
        airflow_db_schema: Optional[str] = None,
        airflow_db_host: Optional[str] = None,
        airflow_db_port: Optional[str] = None,
        airflow_db_driver: str = "postgresql+psycopg2",
        # Airflow db connections in the format { conn_id: conn_url },
        # converted to env var: AIRFLOW_CONN__conn_id = conn_url,
        db_connections: Optional[Dict] = None,
        # Configure airflow scheduler,
        # Init Airflow scheduler as a deamon process,
        init_airflow_scheduler: bool = False,
        # Configure airflow webserver,
        # Init Airflow webserver when the container starts,
        init_airflow_webserver: bool = False,
        # webserver port on the container,
        airflow_webserver_container_port: int = 8080,
        # webserver port on the host machine,
        airflow_webserver_host_port: int = 8080,
        # webserver port name on K8sContainer,
        airflow_webserver_port_name: str = "http",
        # Configure the container,
        container_name: Optional[str] = None,
        # Only used by the DockerContainer,
        container_detach: bool = True,
        container_auto_remove: bool = True,
        container_remove: bool = True,
        # Add env variables to container env,
        env: Optional[Dict[str, str]] = None,
        # Read env variables from a file in yaml format,
        env_file: Optional[Path] = None,
        # Configure the ConfigMap used for env variables that are not Secret,
        config_map_name: Optional[str] = None,
        # Configure the Secret used for env variables that are Secret,
        secret_name: Optional[str] = None,
        # Read secrets from a file in yaml format,
        secrets_file: Optional[Path] = None,
        # Configure the databox deploy,
        deploy_name: Optional[str] = None,
        pod_name: Optional[str] = None,
        # Other args,
        load_examples: bool = False,
        print_env_on_load: bool = True,
        dev_mode: Optional[DevboxDevModeArgs] = None,
        # If True, skip resource creation if active resources with the same name exist.
        use_cache: bool = True,
        # If True, log extra debug messages
        use_verbose_logs: bool = False,
    ):
        super().__init__()
        try:
            self.args: DevboxArgs = DevboxArgs(
                name=name,
                version=version,
                enabled=enabled,
                image_name=image_name,
                image_tag=image_tag,
                entrypoint=entrypoint,
                command=command,
                mount_workspace=mount_workspace,
                workspace_volume_name=workspace_volume_name,
                workspace_parent_container_path=workspace_parent_container_path,
                install_requirements=install_requirements,
                requirements_file_path=requirements_file_path,
                mount_aws_config=mount_aws_config,
                aws_config_path=aws_config_path,
                aws_config_container_path=aws_config_container_path,
                init_airflow=init_airflow,
                use_products_as_airflow_dags=use_products_as_airflow_dags,
                airflow_dags_path=airflow_dags_path,
                create_airflow_test_user=create_airflow_test_user,
                airflow_executor=airflow_executor,
                mount_airflow_home=mount_airflow_home,
                airflow_home_dir=airflow_home_dir,
                airflow_home_container_path=airflow_home_container_path,
                init_airflow_db=init_airflow_db,
                wait_for_airflow_db=wait_for_airflow_db,
                airflow_db_app=airflow_db_app,
                airflow_db_user=airflow_db_user,
                airflow_db_password=airflow_db_password,
                airflow_db_schema=airflow_db_schema,
                airflow_db_host=airflow_db_host,
                airflow_db_port=airflow_db_port,
                airflow_db_driver=airflow_db_driver,
                db_connections=db_connections,
                init_airflow_scheduler=init_airflow_scheduler,
                init_airflow_webserver=init_airflow_webserver,
                airflow_webserver_container_port=airflow_webserver_container_port,
                airflow_webserver_host_port=airflow_webserver_host_port,
                airflow_webserver_port_name=airflow_webserver_port_name,
                container_name=container_name,
                container_detach=container_detach,
                container_auto_remove=container_auto_remove,
                container_remove=container_remove,
                env=env,
                env_file=env_file,
                config_map_name=config_map_name,
                secret_name=secret_name,
                secrets_file=secrets_file,
                deploy_name=deploy_name,
                pod_name=pod_name,
                load_examples=load_examples,
                print_env_on_load=print_env_on_load,
                dev_mode=dev_mode,
                use_cache=use_cache,
                use_verbose_logs=use_verbose_logs,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    def get_container_name(self) -> str:
        return self.args.container_name or get_default_container_name(self.args.name)

    def get_env_data_from_file(self) -> Optional[Dict[str, str]]:
        import yaml

        env_file_path = self.args.env_file
        if (
            env_file_path is not None
            and env_file_path.exists()
            and env_file_path.is_file()
        ):
            # logger.debug(f"Reading {env_file_path}")
            env_data_from_file = yaml.safe_load(env_file_path.read_text())
            if env_data_from_file is not None and isinstance(env_data_from_file, dict):
                return env_data_from_file
            else:
                print_error(f"Invalid env_file: {env_file_path}")
        return None

    def get_secret_data_from_file(self) -> Optional[Dict[str, str]]:
        import yaml

        secrets_file_path = self.args.secrets_file
        if (
            secrets_file_path is not None
            and secrets_file_path.exists()
            and secrets_file_path.is_file()
        ):
            # logger.debug(f"Reading {secrets_file_path}")
            secret_data_from_file = yaml.safe_load(secrets_file_path.read_text())
            if secret_data_from_file is not None and isinstance(
                secret_data_from_file, dict
            ):
                return secret_data_from_file
            else:
                print_error(f"Invalid secrets_file: {secrets_file_path}")
        return None

    ######################################################
    ## Docker Resources
    ######################################################

    def init_airflow_on_docker_container(self, container: DockerContainer) -> None:
        """
        Initialize airflow on a docker container
        """

        if not self.args.init_airflow:
            return

        # Update init_airflow_db arg if None
        if self.args.init_airflow_db is None:
            self.args.init_airflow_db = True

        # Workspace paths
        workspace_name = self.workspace_root_path.stem
        workspace_root_container_path = Path(
            self.args.workspace_parent_container_path
        ).joinpath(workspace_name)
        products_dir_container_path = workspace_root_container_path.joinpath(
            self.products_dir
        )

        # Airflow db connection
        airflow_db_user = self.args.airflow_db_user
        airflow_db_password = self.args.airflow_db_password
        airflow_db_schema = self.args.airflow_db_schema
        airflow_db_host = self.args.airflow_db_host
        airflow_db_port = self.args.airflow_db_port
        airflow_db_driver = self.args.airflow_db_driver
        if self.args.airflow_db_app is not None and isinstance(
            self.args.airflow_db_app, DbApp
        ):
            logger.debug(
                f"Reading Airflow db connection from DbApp: {self.args.airflow_db_app.name}"
            )
            if airflow_db_user is None:
                airflow_db_user = self.args.airflow_db_app.get_db_user()
            if airflow_db_password is None:
                airflow_db_password = self.args.airflow_db_app.get_db_password()
            if airflow_db_schema is None:
                airflow_db_schema = self.args.airflow_db_app.get_db_schema()
            if airflow_db_host is None:
                airflow_db_host = self.args.airflow_db_app.get_db_host_docker()
            if airflow_db_port is None:
                airflow_db_port = self.args.airflow_db_app.get_db_port_docker()
            if airflow_db_driver is None:
                airflow_db_driver = self.args.airflow_db_app.get_db_driver()
        db_connection_url = f"{airflow_db_driver}://{airflow_db_user}:{airflow_db_password}@{airflow_db_host}:{airflow_db_port}/{airflow_db_schema}"

        airflow_env: Dict[str, Any] = {
            "INIT_AIRFLOW": str(self.args.init_airflow),
            "INIT_AIRFLOW_DB": str(self.args.init_airflow_db),
            "WAIT_FOR_AIRFLOW_DB": str(self.args.wait_for_airflow_db),
            "AIRFLOW_DB_USER": str(airflow_db_user),
            "AIRFLOW_DB_PASSWORD": str(airflow_db_password),
            "AIRFLOW_SCHEMA": str(airflow_db_schema),
            "AIRFLOW_DB_HOST": str(airflow_db_host),
            "AIRFLOW_DB_PORT": str(airflow_db_port),
            "INIT_AIRFLOW_SCHEDULER": str(self.args.init_airflow_scheduler),
            "INIT_AIRFLOW_WEBSERVER": str(self.args.init_airflow_webserver),
            "AIRFLOW__CORE__LOAD_EXAMPLES": str(self.args.load_examples),
            "CREATE_AIRFLOW_TEST_USER": str(self.args.create_airflow_test_user),
            "AIRFLOW__CORE__EXECUTOR": str(self.args.airflow_executor),
        }
        # Set the AIRFLOW__CORE__SQL_ALCHEMY_CONN
        if "None" not in db_connection_url:
            airflow_env["AIRFLOW__CORE__SQL_ALCHEMY_CONN"] = db_connection_url
        # Set the AIRFLOW__CORE__DAGS_FOLDER
        if self.args.mount_workspace and self.args.use_products_as_airflow_dags:
            airflow_env["AIRFLOW__CORE__DAGS_FOLDER"] = str(products_dir_container_path)
        elif self.args.airflow_dags_path is not None:
            airflow_env["AIRFLOW__CORE__DAGS_FOLDER"] = self.args.airflow_dags_path
        # Set the AIRFLOW__CONN_ variables
        if self.args.db_connections is not None:
            for conn_id, conn_url in self.args.db_connections.items():
                try:
                    af_conn_id = str("AIRFLOW_CONN_{}".format(conn_id)).upper()
                    airflow_env[af_conn_id] = conn_url
                except Exception as e:
                    logger.exception(e)
                    continue
        # Update the container.environment to include airflow_env
        container.environment.update(airflow_env)

        # Open the airflow webserver port if init_airflow_webserver = True
        if self.args.init_airflow_webserver:
            ws_port: Dict[str, int] = {
                str(
                    self.args.airflow_webserver_container_port
                ): self.args.airflow_webserver_host_port,
            }
            if container.ports is None:
                container.ports = {}
            container.ports.update(ws_port)

    def get_docker_rg(
        self, docker_build_context: DockerBuildContext
    ) -> Optional[DockerResourceGroup]:

        app_name = self.args.name
        logger.debug(f"Building {app_name} DockerResourceGroup")

        # Workspace paths
        workspace_name = self.workspace_root_path.stem
        workspace_root_container_path = Path(
            self.args.workspace_parent_container_path
        ).joinpath(workspace_name)
        requirements_file_container_path = workspace_root_container_path.joinpath(
            self.args.requirements_file_path
        )
        scripts_dir_container_path = workspace_root_container_path.joinpath(
            self.scripts_dir
        )
        storage_dir_container_path = workspace_root_container_path.joinpath(
            self.storage_dir
        )
        meta_dir_container_path = workspace_root_container_path.joinpath(self.meta_dir)
        products_dir_container_path = workspace_root_container_path.joinpath(
            self.products_dir
        )
        notebooks_dir_container_path = workspace_root_container_path.joinpath(
            self.notebooks_dir
        )
        workspace_config_dir_container_path = workspace_root_container_path.joinpath(
            self.workspace_config_dir
        )

        # Container Environment
        container_env: Dict[str, Any] = {
            # Env variables used by data workflows and data assets
            "PHI_WORKSPACE_PARENT": str(self.args.workspace_parent_container_path),
            "PHI_WORKSPACE_ROOT": str(workspace_root_container_path),
            "PHI_SCRIPTS_DIR": str(scripts_dir_container_path),
            "PHI_STORAGE_DIR": str(storage_dir_container_path),
            "PHI_META_DIR": str(meta_dir_container_path),
            "PHI_PRODUCTS_DIR": str(products_dir_container_path),
            "PHI_NOTEBOOKS_DIR": str(notebooks_dir_container_path),
            "PHI_WORKSPACE_CONFIG_DIR": str(workspace_config_dir_container_path),
            "INSTALL_REQUIREMENTS": str(self.args.install_requirements),
            "REQUIREMENTS_FILE_PATH": str(requirements_file_container_path),
            # Print env when the container starts
            "PRINT_ENV_ON_LOAD": str(self.args.print_env_on_load),
            # Dev mode args
            "INSTALL_PHIDATA_DEV": str(self.args.dev_mode.install_phidata_dev)
            if self.args.dev_mode is not None
            else str(False),
            "PHIDATA_DIR_PATH": self.args.dev_mode.phidata_dir_container_path
            if self.args.dev_mode is not None
            else str(False),
        }
        # Update the container env using env_file
        env_data_from_file = self.get_env_data_from_file()
        if env_data_from_file is not None:
            container_env.update(env_data_from_file)
        # Update the container env with user provided env
        if self.args.env is not None and isinstance(self.args.env, dict):
            container_env.update(self.args.env)

        # Container Volumes
        # container_volumes is a dictionary which configures the volumes to mount
        # inside the container. The key is either the host path or a volume name,
        # and the value is a dictionary with 2 keys:
        #   bind - The path to mount the volume inside the container
        #   mode - Either rw to mount the volume read/write, or ro to mount it read-only.
        # For example:
        # {
        #   '/home/user1/': {'bind': '/mnt/vol2', 'mode': 'rw'},
        #   '/var/www': {'bind': '/mnt/vol1', 'mode': 'ro'}
        # }
        container_volumes = {}
        # Create a volume for the workspace dir
        if self.args.mount_workspace:
            workspace_root_path_str = str(self.workspace_root_path)
            workspace_root_container_path_str = str(workspace_root_container_path)
            logger.debug(f"Mounting: {workspace_root_path_str}")
            logger.debug(f"\tto: {workspace_root_container_path_str}")
            container_volumes[workspace_root_path_str] = {
                "bind": workspace_root_container_path_str,
                "mode": "rw",
            }
        # Create a volume for aws credentials
        if self.args.mount_aws_config:
            aws_config_path_str = str(self.args.aws_config_path)
            logger.debug(f"Mounting: {aws_config_path_str}")
            logger.debug(f"\tto: {self.args.aws_config_container_path}")
            container_volumes[aws_config_path_str] = {
                "bind": self.args.aws_config_container_path,
                "mode": "ro",
            }
            container_env[
                "AWS_CONFIG_FILE"
            ] = f"{self.args.aws_config_container_path}/config"
            container_env[
                "AWS_SHARED_CREDENTIALS_FILE"
            ] = f"{self.args.aws_config_container_path}/credentials"
        # Create a volume for Phidata dev mode if dev_mode is enabled
        if self.args.dev_mode is not None:
            # mount the phidata directory if install_phidata_dev = True
            if (
                self.args.dev_mode.install_phidata_dev
                and self.args.dev_mode.phidata_dir_path is not None
            ):
                phidata_dir_absolute_path_str = str(self.args.dev_mode.phidata_dir_path)
                logger.debug(f"Mounting: {phidata_dir_absolute_path_str}")
                logger.debug(f"\tto: {self.args.dev_mode.phidata_dir_container_path}")
                container_volumes[phidata_dir_absolute_path_str] = {
                    "bind": self.args.dev_mode.phidata_dir_container_path,
                    "mode": "rw",
                }
        # Mount airflow home if needed (experimental)
        if self.args.mount_airflow_home is not None:
            airflow_home_path_str = str(
                self.workspace_root_path.joinpath(self.args.airflow_home_dir)
            )
            airflow_home_container_path_str = str(self.args.airflow_home_container_path)
            logger.debug(f"Mounting: {airflow_home_path_str}")
            logger.debug(f"\tto: {airflow_home_container_path_str}")
            container_volumes[airflow_home_path_str] = {
                "bind": airflow_home_container_path_str,
                "mode": "rw",
            }

        # Create the container
        docker_container = DockerContainer(
            name=self.get_container_name(),
            image=get_image_str(self.args.image_name, self.args.image_tag),
            entrypoint=self.args.entrypoint,
            command=self.args.command,
            detach=self.args.container_detach,
            auto_remove=self.args.container_auto_remove,
            remove=self.args.container_remove,
            stdin_open=True,
            tty=True,
            environment=container_env,
            network=docker_build_context.network,
            volumes=container_volumes,
            use_cache=self.args.use_cache,
            use_verbose_logs=self.args.use_verbose_logs,
        )
        # Initialize airflow on container
        self.init_airflow_on_docker_container(docker_container)
        # logger.debug(f"Devbox Container Env: {docker_container.environment}")

        docker_rg = DockerResourceGroup(
            name=app_name,
            enabled=self.args.enabled,
            network=DockerNetwork(name=docker_build_context.network),
            containers=[docker_container],
        )
        return docker_rg

    def init_docker_resource_groups(
        self, docker_build_context: DockerBuildContext
    ) -> None:
        docker_rg = self.get_docker_rg(docker_build_context)
        if docker_rg is not None:
            if self.docker_resource_groups is None:
                self.docker_resource_groups = OrderedDict()
            self.docker_resource_groups[docker_rg.name] = docker_rg

    ######################################################
    ## K8s Resources
    ######################################################

    def init_airflow_on_k8s_container(
        self, container: CreateContainer, k8s_resource_group: CreateK8sResourceGroup
    ) -> None:
        """
        Initialize airflow on a k8s container
        """

        if not self.args.init_airflow:
            return

        # Update init_airflow_db arg if None
        if self.args.init_airflow_db is None:
            self.args.init_airflow_db = True

        airflow_cm = CreateConfigMap(
            cm_name=get_default_configmap_name("devbox-airflow"),
            app_name=self.args.name,
            data={
                "INIT_AIRFLOW": str(self.args.init_airflow),
                "INIT_AIRFLOW_DB": str(self.args.init_airflow_db),
                "WAIT_FOR_AIRFLOW_DB": str(self.args.wait_for_airflow_db),
                "AIRFLOW_DB_CONN_URL": self.args.airflow_db_conn_url,
                "AIRFLOW_DB_CONN_PORT": self.args.airflow_db_conn_port,
                "AIRFLOW_DB_USER": self.args.airflow_db_user,
                "AIRFLOW_DB_PASSWORD": self.args.airflow_db_password,
                "AIRFLOW_SCHEMA": self.args.airflow_schema,
                "INIT_AIRFLOW_SCHEDULER": str(self.args.init_airflow_scheduler),
                "INIT_AIRFLOW_WEBSERVER": str(self.args.init_airflow_webserver),
                "CREATE_AIRFLOW_TEST_USER": str(self.args.init_airflow_webserver),
                "AIRFLOW__CORE__LOAD_EXAMPLES": str(self.args.load_examples),
            },
        )

        # Add airflow_cm to container env and
        if container.envs_from_configmap is None:
            container.envs_from_configmap = []
        container.envs_from_configmap.append(airflow_cm.cm_name)
        if k8s_resource_group.config_maps is None:
            k8s_resource_group.config_maps = []
        k8s_resource_group.config_maps.append(airflow_cm)

        # Set the AIRFLOW__CORE__DAGS_FOLDER
        if self.args.mount_workspace and self.args.use_products_as_airflow_dags:
            projects_dir_container_path = Path(
                self.args.workspace_parent_container_path
            ).joinpath(self.args.workspace_dir_path.stem, self.args.projects_dir)
            airflow_cm.data["AIRFLOW__CORE__DAGS_FOLDER"] = projects_dir_container_path
        elif self.args.airflow_dags_path is not None:
            airflow_cm.data["AIRFLOW__CORE__DAGS_FOLDER"] = self.args.airflow_dags_path
        if self.args.db_connections is not None:
            for conn_id, conn_url in self.args.db_connections.items():
                try:
                    af_conn_id = str("AIRFLOW_CONN_{}".format(conn_id)).upper()
                    airflow_cm.data[af_conn_id] = conn_url
                except Exception as e:
                    logger.exception(e)
                    continue
        # Open the airflow webserver port if init_airflow_webserver = True
        if self.args.init_airflow_webserver:
            ws_port = CreatePort(
                name=self.args.airflow_webserver_port_name,
                container_port=self.args.airflow_webserver_container_port,
            )
            if container.ports is None:
                container.ports = []
            container.ports.append(ws_port)

    def get_devbox_k8s_rg(
        self, k8s_build_context: K8sBuildContext
    ) -> Optional[K8sResourceGroup]:
        logger.debug(f"Init Devbox K8sResourceGroup")

        # Set the Devbox Env
        # Devbox Config-Map is a dict of environment variables which are not secrets.
        # The keys of this configmap will become env variables in the devbox container
        workspace_dir_container_path = Path(
            self.args.workspace_parent_container_path
        ).joinpath(self.args.workspace_dir_path.stem)
        projects_dir_container_path = workspace_dir_container_path.joinpath(
            self.args.projects_dir
        )
        devbox_cm = CreateConfigMap(
            cm_name=get_default_configmap_name(self.args.name),
            app_name=self.args.name,
            data={
                # Env variables used by data workflows and data assets
                "PHI_DATA_DIR": self.args.data_dir_container_path,
                "PHI_PROJECTS_DIR": str(projects_dir_container_path),
                "PHI_WORKSPACE_DIR": str(workspace_dir_container_path),
                "PHI_WORKSPACE_PARENT": str(self.args.workspace_parent_container_path),
                # Configure python packages to install
                "INSTALL_REQUIREMENTS": str(self.args.install_requirements),
                "REQUIREMENTS_FILE_PATH": self.args.requirements_file_container_path,
                # Print these env variables when the container is loaded
                "PRINT_ENV_ON_LOAD": str(self.args.print_env_on_load),
                # Dev mode args
                "INSTALL_PHIDATA_DEV": str(self.args.dev_mode.install_phidata_dev)
                if self.args.dev_mode is not None
                else str(False),
                "PHIDATA_DIR_PATH": self.args.dev_mode.phidata_dir_container_path
                if self.args.dev_mode is not None
                else str(False),
            },
        )
        # Update the devbox env with user provided env
        if self.args.env is not None and isinstance(self.args.env, dict):
            devbox_cm.data.update(self.args.env)

        # Devbox Volumes
        devbox_volumes = []
        # Create a volume for the requirements file
        # if install_requirements = True
        if self.args.install_requirements:
            requirements_file_absolute_path = str(
                self.args.workspace_dir_path.joinpath(self.args.requirements_file_path)
            )
            logger.debug(f"Mounting: {requirements_file_absolute_path}")
            logger.debug(f"\tto: {self.args.requirements_file_container_path}")
            requirements_volume = CreateVolume(
                volume_name=self.args.requirements_volume_name,
                app_name=self.args.name,
                mount_path=self.args.requirements_file_container_path,
                volume_type=VolumeType.HOST_PATH,
                host_path=HostPathVolumeSource(path=requirements_file_absolute_path),
            )
            devbox_volumes.append(requirements_volume)
        # Create a volume for the data dir
        # if create_data_volume = True
        if self.args.create_data_volume:
            if (
                self.args.use_host_path_for_data_volume
                and self.args.data_dir_path is not None
            ):
                data_dir_absolute_path = str(
                    self.args.workspace_dir_path.joinpath(self.args.data_dir_path)
                )
                logger.debug(f"Mounting: {data_dir_absolute_path}")
                logger.debug(f"\tto: {self.args.data_dir_container_path}")
                data_volume = CreateVolume(
                    volume_name=self.args.data_dir_volume_name,
                    app_name=self.args.name,
                    mount_path=self.args.data_dir_container_path,
                    volume_type=VolumeType.HOST_PATH,
                    host_path=HostPathVolumeSource(
                        path=data_dir_absolute_path,
                    ),
                )
                devbox_volumes.append(data_volume)
            elif self.args.use_empty_dir_for_data_volume:
                logger.debug(f"Mounting EmptyDir")
                logger.debug(f"\tto: {self.args.data_dir_container_path}")
                data_volume = CreateVolume(
                    volume_name=self.args.data_dir_volume_name,
                    app_name=self.args.name,
                    mount_path=self.args.data_dir_container_path,
                    volume_type=VolumeType.EMPTY_DIR,
                    empty_dir=EmptyDirVolumeSource(),
                )
                devbox_volumes.append(data_volume)
        # Create a volume for the projects dir
        # if mount_workspace = True
        if self.args.mount_workspace:
            if self.args.use_host_path_for_workspace:
                ws_dir_container_path_str = str(
                    Path(self.args.workspace_parent_container_path).joinpath(
                        self.args.workspace_dir_path.stem
                    )
                )
                ws_dir_path_str = str(self.args.workspace_dir_path)
                logger.debug(f"Mounting: {ws_dir_path_str}")
                logger.debug(f"\tto: {ws_dir_container_path_str}")
                workspace_volume = CreateVolume(
                    volume_name=self.args.workspace_volume_name,
                    app_name=self.args.name,
                    mount_path=ws_dir_container_path_str,
                    volume_type=VolumeType.HOST_PATH,
                    host_path=HostPathVolumeSource(
                        path=ws_dir_path_str,
                    ),
                )
                devbox_volumes.append(workspace_volume)
        # Airflow conf file
        if self.args.mount_airflow_dir:
            airflow_dir_file_absolute_path = str(
                self.args.workspace_dir_path.joinpath(self.args.airflow_dir_path)
            )
            logger.debug(f"Mounting: {airflow_dir_file_absolute_path}")
            logger.debug(f"\tto: {self.args.airflow_dir_container_path}")
            airflow_dir_volume = CreateVolume(
                volume_name=self.args.airflow_volume_name,
                app_name=self.args.name,
                mount_path=self.args.airflow_dir_container_path,
                volume_type=VolumeType.HOST_PATH,
                host_path=HostPathVolumeSource(path=airflow_dir_file_absolute_path),
            )
            devbox_volumes.append(airflow_dir_volume)
        # Create a volume for Phidata dev mode
        # if dev_mode is enabled
        if self.args.dev_mode is not None:
            # mount the phidata directory if install_phidata_dev = True
            if (
                self.args.dev_mode.install_phidata_dev
                and self.args.dev_mode.phidata_dir_path is not None
            ):
                phidata_dir_absolute_path = str(self.args.dev_mode.phidata_dir_path)
                logger.debug(f"Mounting: {phidata_dir_absolute_path}")
                logger.debug(f"\tto: {self.args.dev_mode.phidata_dir_container_path}")
                phidata_volume = CreateVolume(
                    volume_name=self.args.dev_mode.phidata_volume_name,
                    app_name=self.args.name,
                    mount_path=self.args.dev_mode.phidata_dir_container_path,
                    volume_type=VolumeType.HOST_PATH,
                    host_path=HostPathVolumeSource(
                        path=phidata_dir_absolute_path,
                    ),
                )
                devbox_volumes.append(phidata_volume)

        # Create the devbox container
        devbox_k8s_container = CreateContainer(
            container_name=f"{self.args.name}-container",
            app_name=self.args.name,
            image_name=self.args.image_name,
            image_tag=self.args.image_tag,
            # Equivalent to docker images CMD
            args=self.args.command,
            # Equivalent to docker images ENTRYPOINT
            command=self.args.entrypoint,
            envs_from_configmap=[devbox_cm.cm_name],
            volumes=devbox_volumes,
        )
        devbox_deployment = CreateDeployment(
            deploy_name=f"{self.args.name}-deploy",
            pod_name=f"{self.args.name}-pod",
            app_name=self.args.name,
            namespace=k8s_build_context.namespace,
            service_account_name=k8s_build_context.service_account_name,
            containers=[devbox_k8s_container],
            volumes=devbox_volumes,
            labels=k8s_build_context.labels,
        )
        devbox_k8s_resource_group = CreateK8sResourceGroup(
            name=self.args.name,
            enabled=self.args.enabled,
            config_maps=[devbox_cm],
            deployments=[devbox_deployment],
        )
        # Initialize airflow on container
        self.init_airflow_on_k8s_container(
            devbox_k8s_container, devbox_k8s_resource_group
        )
        return devbox_k8s_resource_group.create()

    def init_k8s_resource_groups(self, k8s_build_context: K8sBuildContext) -> None:
        devbox_k8s_rg: Optional[K8sResourceGroup] = self.get_devbox_k8s_rg(
            k8s_build_context
        )
        # logger.debug("devbox_k8s_rg: {}".format(devbox_k8s_rg))
        if devbox_k8s_rg is not None:
            if self.k8s_resource_groups is None:
                self.k8s_resource_groups = OrderedDict()
            self.k8s_resource_groups[devbox_k8s_rg.name] = devbox_k8s_rg



from typing import Dict, List, Optional, Type

from pydantic import BaseModel

from phidata.infra.k8s.create.apps.v1.deployment import CreateDeployment
from phidata.infra.k8s.create.core.v1.config_map import CreateConfigMap
from phidata.infra.k8s.create.core.v1.namespace import CreateNamespace
from phidata.infra.k8s.create.core.v1.secret import CreateSecret
from phidata.infra.k8s.create.core.v1.service import CreateService
from phidata.infra.k8s.create.core.v1.service_account import CreateServiceAccount
from phidata.infra.k8s.create.apiextensions_k8s_io.v1.custom_resource_definition import (
    CreateCustomResourceDefinition,
)
from phidata.infra.k8s.create.rbac_authorization_k8s_io.v1.cluster_role import (
    CreateClusterRole,
)
from phidata.infra.k8s.create.rbac_authorization_k8s_io.v1.cluste_role_binding import (
    CreateClusterRoleBinding,
)
from phidata.infra.k8s.create.apiextensions_k8s_io.v1.custom_object import (
    CreateCustomObject,
)
from phidata.infra.k8s.create.storage_k8s_io.v1.storage_class import CreateStorageClass
from phidata.infra.k8s.resource.group import (
    K8sResourceGroup,
    K8sBuildContext,
)
from phidata.utils.cli_console import print_error
from phidata.utils.log import logger


class CreateK8sResourceGroup(BaseModel):
    name: str
    enabled: bool = True
    weight: int = 100
    ns: Optional[CreateNamespace] = None
    sa: Optional[CreateServiceAccount] = None
    cr: Optional[CreateClusterRole] = None
    crb: Optional[CreateClusterRoleBinding] = None
    secrets: Optional[List[CreateSecret]] = None
    config_maps: Optional[List[CreateConfigMap]] = None
    storage_classes: Optional[List[CreateStorageClass]] = None
    services: Optional[List[CreateService]] = None
    deployments: Optional[List[CreateDeployment]] = None
    custom_objects: Optional[List[CreateCustomObject]] = None
    crds: Optional[List[CreateCustomResourceDefinition]] = None

    def create(self) -> Optional[K8sResourceGroup]:
        """Creates the K8sResourceGroup"""

        # logger.debug(f"Creating K8sResourceGroup: {self.name}")
        k8s_resource_group: K8sResourceGroup = K8sResourceGroup(
            name=self.name,
            enabled=self.enabled,
            weight=self.weight,
        )

        for key, value in self.__dict__.items():

            if value is None:
                continue
            # logger.debug(f"Parsing {key}")

            ######################################################
            ## Create Namespace
            ######################################################
            elif key == "ns":
                from phidata.infra.k8s.resource.core.v1.namespace import Namespace

                if not isinstance(value, CreateNamespace):
                    logger.error(
                        f"Expected: CreateNamespace. Received: {type(value)}. Skipping."
                    )
                    continue

                _ns_resource: Optional[Namespace] = value.create()
                if _ns_resource is not None:
                    k8s_resource_group.ns = _ns_resource

            ######################################################
            ## Create ServiceAccount
            ######################################################
            elif key == "sa":
                from phidata.infra.k8s.resource.core.v1.service_account import (
                    ServiceAccount,
                )

                if not isinstance(value, CreateServiceAccount):
                    logger.error(
                        f"Expected: CreateServiceAccount. Received: {type(value)}. Skipping."
                    )
                    continue

                _sa_resource: Optional[ServiceAccount] = value.create()
                if _sa_resource is not None:
                    k8s_resource_group.sa = _sa_resource

            ######################################################
            ## Create ClusterRole
            ######################################################
            elif key == "cr":
                from phidata.infra.k8s.resource.rbac_authorization_k8s_io.v1.cluster_role import (
                    ClusterRole,
                )

                if not isinstance(value, CreateClusterRole):
                    logger.error(
                        f"Expected: CreateClusterRole. Received: {type(value)}. Skipping."
                    )
                    continue

                _cr_resource: Optional[ClusterRole] = value.create()
                if _cr_resource is not None:
                    k8s_resource_group.cr = _cr_resource

            ######################################################
            ## Create ClusterRoleBinding
            ######################################################
            elif key == "crb":
                from phidata.infra.k8s.resource.rbac_authorization_k8s_io.v1.cluste_role_binding import (
                    ClusterRoleBinding,
                )

                if not isinstance(value, CreateClusterRoleBinding):
                    logger.error(
                        f"Expected: CreateClusterRoleBinding. Received: {type(value)}. Skipping."
                    )
                    continue

                _crb_resource: Optional[ClusterRoleBinding] = value.create()
                if _crb_resource is not None:
                    k8s_resource_group.crb = _crb_resource

            ######################################################
            ## Create Secrets
            ######################################################
            elif key == "secrets":
                if not isinstance(value, List):
                    logger.error(
                        f"Expected: List[CreateSecret]. Received: {type(value)}. Skipping."
                    )
                    continue

                from phidata.infra.k8s.resource.core.v1.secret import Secret

                secret_resources: List[Secret] = []
                for _secret in value:
                    if not isinstance(_secret, CreateSecret):
                        logger.error(
                            f"Expected: CreateSecret. Received: {type(_secret)}. Skipping."
                        )
                        continue
                    _secret_resource: Optional[Secret] = _secret.create()
                    if _secret_resource is not None:
                        secret_resources.append(_secret_resource)
                if len(secret_resources) >= 0:
                    k8s_resource_group.secrets = secret_resources

            ######################################################
            ## Create ConfigMaps
            ######################################################
            elif key == "config_maps":
                if not isinstance(value, List):
                    logger.error(
                        f"Expected: List[CreateConfigMap]. Received: {type(value)}. Skipping."
                    )
                    continue

                from phidata.infra.k8s.resource.core.v1.config_map import ConfigMap

                cm_resources: List[ConfigMap] = []
                for _config_map in value:
                    if not isinstance(_config_map, CreateConfigMap):
                        logger.error(
                            f"Expected: CreateConfigMap. Received: {type(_config_map)}. Skipping."
                        )
                        continue
                    _cm_resource: Optional[ConfigMap] = _config_map.create()
                    if _cm_resource is not None:
                        cm_resources.append(_cm_resource)
                if len(cm_resources) >= 0:
                    k8s_resource_group.config_maps = cm_resources

            ######################################################
            ## Create StorageClasses
            ######################################################
            elif key == "storage_classes":
                if not isinstance(value, List):
                    logger.error(
                        f"Expected: List[CreateStorageClass]. Received: {type(value)}. Skipping."
                    )
                    continue

                from phidata.infra.k8s.resource.storage_k8s_io.v1.storage_class import (
                    StorageClass,
                )

                sc_resources: List[StorageClass] = []
                for _storage_class in value:
                    if not isinstance(_storage_class, CreateStorageClass):
                        logger.error(
                            f"Expected: CreateStorageClass. Received: {type(_storage_class)}. Skipping."
                        )
                        continue
                    _sc_resource: Optional[StorageClass] = _storage_class.create()
                    if _sc_resource is not None:
                        sc_resources.append(_sc_resource)
                if len(sc_resources) >= 0:
                    k8s_resource_group.storage_classes = sc_resources

            ######################################################
            ## Create Services
            ######################################################
            if key == "services":
                if not isinstance(value, List):
                    logger.error(
                        f"Expected: List[CreateService]. Received: {type(value)}. Skipping."
                    )
                    continue

                # Add necessary imports here to speed up file load time
                from phidata.infra.k8s.resource.core.v1.service import Service

                service_resources: List[Service] = []
                for _service in value:
                    if not isinstance(_service, CreateService):
                        logger.error(
                            f"Expected: CreateService. Received: {type(_service)}. Skipping."
                        )
                        continue
                    _service_resource: Optional[Service] = _service.create()
                    if _service_resource is not None:
                        service_resources.append(_service_resource)
                if len(service_resources) >= 0:
                    k8s_resource_group.services = service_resources

            ######################################################
            ## Create Deployments
            ######################################################
            if key == "deployments":
                if not isinstance(value, List):
                    logger.error(
                        f"Expected: List[CreateDeployment]. Received: {type(value)}. Skipping."
                    )
                    continue

                # Add necessary imports here to speed up file load time
                from phidata.infra.k8s.resource.apps.v1.deployment import Deployment

                deploy_resources: List[Deployment] = []
                for _deployment in value:
                    if not isinstance(_deployment, CreateDeployment):
                        logger.error(
                            f"Expected: CreateDeployment. Received: {type(_deployment)}. Skipping."
                        )
                        continue
                    _deployment_resource: Optional[Deployment] = _deployment.create()
                    if _deployment_resource is not None:
                        deploy_resources.append(_deployment_resource)
                # logger.debug(deploy_resources)
                if len(deploy_resources) >= 0:
                    k8s_resource_group.deployments = deploy_resources

            ######################################################
            ## Create CustomResourceDefinitions
            ######################################################
            if key == "custom_objects":
                if not isinstance(value, List):
                    logger.error(
                        f"Expected: List[CreateCustomObject]. Received: {type(value)}. Skipping."
                    )
                    continue

                # Add necessary imports here to speed up file load time
                from phidata.infra.k8s.resource.apiextensions_k8s_io.v1.custom_object import (
                    CustomObject,
                )

                custom_object_resources: List[CustomObject] = []
                for _custom_object in value:
                    if not isinstance(_custom_object, CreateCustomObject):
                        logger.error(
                            f"Expected: CreateCustomObject. Received: {type(_custom_object)}. Skipping."
                        )
                        continue
                    _custom_object_resource: Optional[
                        CustomObject
                    ] = _custom_object.create()
                    if _custom_object_resource is not None:
                        custom_object_resources.append(_custom_object_resource)
                if len(custom_object_resources) >= 0:
                    k8s_resource_group.custom_objects = custom_object_resources

            ######################################################
            ## Create CustomResourceDefinitions
            ######################################################
            if key == "crds":
                if not isinstance(value, List):
                    logger.error(
                        f"Expected: List[CreateCustomResourceDefinition]. Received: {type(value)}. Skipping."
                    )
                    continue

                # Add necessary imports here to speed up file load time
                from phidata.infra.k8s.resource.apiextensions_k8s_io.v1.custom_resource_definition import (
                    CustomResourceDefinition,
                )

                crd_resources: List[CustomResourceDefinition] = []
                for _crd in value:
                    if not isinstance(_crd, CreateCustomResourceDefinition):
                        logger.error(
                            f"Expected: CreateCustomResourceDefinition. Received: {type(_crd)}. Skipping."
                        )
                        continue
                    _crd_resource: Optional[CustomResourceDefinition] = _crd.create()
                    if _crd_resource is not None:
                        crd_resources.append(_crd_resource)
                if len(crd_resources) >= 0:
                    k8s_resource_group.crds = crd_resources

        logger.debug(f"-*- Initialized K8sResourceGroup for: {self.name}")
        return k8s_resource_group



from collections import defaultdict, OrderedDict
from typing import Any, Dict, List, Optional, Type, Set, cast

from phidata.infra.aws.args import AwsArgs
from phidata.infra.aws.api_client import AwsApiClient
from phidata.infra.aws.exceptions import (
    AwsArgsException,
    AwsApiClientException,
    AwsResourceCreationFailedException,
)
from phidata.infra.aws.resource.base import AwsResource
from phidata.infra.aws.resource.group import AwsResourceGroup
from phidata.infra.aws.resource.types import AwsResourceType
from phidata.infra.aws.resource.utils import filter_and_flatten_aws_resource_group
from phidata.utils.cli_console import (
    print_info,
    print_info,
    print_heading,
    print_subheading,
    print_error,
    confirm_yes_no,
)
from phidata.utils.common import is_empty
from phidata.utils.log import logger


class AwsWorker:
    """This class interacts with the Aws API."""

    def __init__(self, aws_args: AwsArgs) -> None:
        # logger.debug(f"Creating AwsWorker")
        if aws_args is None or not isinstance(aws_args, AwsArgs):
            raise AwsArgsException("AwsArgs invalid: {}".format(aws_args))

        self.aws_args: AwsArgs = aws_args
        self.aws_api_client: AwsApiClient = AwsApiClient(
            aws_region=self.aws_args.aws_region,
            aws_profile=self.aws_args.aws_profile,
        )
        self.aws_resource_group: Optional[AwsResourceGroup] = None
        logger.debug(f"**-+-** AwsWorker created")

    def is_client_initialized(self) -> bool:
        return self.aws_api_client.is_initialized()

    def are_resources_initialized(self) -> bool:
        return True if self.aws_resource_group is not None else False

    ######################################################
    ## Init Resources
    ######################################################

    def init_resources(self) -> None:
        """
        This function populates the self.aws_resource_group dictionary.
        """
        logger.debug("-*- Initializing AwsResources")

        if self.aws_args.resources is not None and isinstance(
            self.aws_args.resources, AwsResourceGroup
        ):
            self.aws_resource_group = self.aws_args.resources
        logger.debug(f"-*- AwsResources built")

    ######################################################
    ## Create Resources
    ######################################################

    def create_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
    ) -> bool:
        logger.debug("-*- Creating AwsResources")
        if self.aws_resource_group is None:
            self.init_resources()
            if self.aws_resource_group is None:
                print_info("No AwsResources available")
                return False

        aws_resources_to_create: Optional[
            List[AwsResourceType]
        ] = filter_and_flatten_aws_resource_group(
            aws_resource_group=self.aws_resource_group,
            name_filter=name_filter,
            type_filter=type_filter,
            sort_order="create",
        )
        if aws_resources_to_create is None or is_empty(aws_resources_to_create):
            print_info("No AwsResources to create")
            return True

        # track the total number of AwsResources to create for validation
        num_resources_to_create: int = len(aws_resources_to_create)
        num_resources_created: int = 0

        # Print the resources for validation
        print_subheading(f"Creating {num_resources_to_create} AwsResources:")
        for rsrc in aws_resources_to_create:
            print_info(f"  -+-> {rsrc.get_resource_type()}: {rsrc.get_resource_name()}")
        confirm = confirm_yes_no("\nConfirm deploy")
        if not confirm:
            print_info("Skipping deploy")
            return False

        for resource in aws_resources_to_create:
            if resource:
                print_info(
                    f"\n-==+==- {resource.get_resource_type()}: {resource.get_resource_name()}"
                )
                # logger.debug(resource)
                try:
                    _resource_created = resource.create(aws_client=self.aws_api_client)
                    if _resource_created:
                        num_resources_created += 1
                except Exception as e:
                    print_error(
                        f"-==+==--> Resource {resource.get_resource_type()}: {resource.get_resource_name()} could not be created."
                    )
                    print_error("Error: {}".format(e))
                    print_error(
                        "Skipping resource creation, please fix and try again..."
                    )

        print_info(
            f"\n# Resources created: {num_resources_created}/{num_resources_to_create}"
        )
        if num_resources_to_create == num_resources_created:
            return True

        print_error(
            f"Resources created: {num_resources_created} do not match Resources required: {num_resources_to_create}"
        )
        return False

    def create_resources_dry_run(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
    ) -> None:

        env = self.aws_args.env
        print_subheading(
            "AwsResources{}".format(f" for env: {env}" if env is not None else "")
        )
        if self.aws_resource_group is None:
            self.init_resources()
            if self.aws_resource_group is None:
                print_info("No AwsResources available")
                return

        aws_resources_to_create: Optional[
            List[AwsResourceType]
        ] = filter_and_flatten_aws_resource_group(
            aws_resource_group=self.aws_resource_group,
            name_filter=name_filter,
            type_filter=type_filter,
            sort_order="create",
        )
        if aws_resources_to_create is None or is_empty(aws_resources_to_create):
            print_info("No AwsResources to create")
            return

        num_resources_to_create: int = len(aws_resources_to_create)
        print_info(
            f"\n-*- Deploying this AwsConfig will create {num_resources_to_create} resources:"
        )
        for resource in aws_resources_to_create:
            print_info(
                f"-==+==- {resource.__class__.__name__}: {resource.get_resource_name()}"
            )

    ######################################################
    ## Read Resources
    ######################################################

    def read_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
    ) -> Dict[str, List[AwsResourceType]]:

        logger.debug("-*- Getting AwsResources")
        if self.aws_resource_group is None:
            self.init_resources()
            if self.aws_resource_group is None:
                print_info("No AwsResources available")
                return {}

    ######################################################
    ## Delete Resources
    ######################################################

    def delete_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
    ) -> bool:

        logger.debug("-*- Deleting AwsResources")
        if self.aws_resource_group is None:
            self.init_resources()
            if self.aws_resource_group is None:
                print_info("No AwsResources available")
                return False

        aws_resources_to_delete: Optional[
            List[AwsResourceType]
        ] = filter_and_flatten_aws_resource_group(
            aws_resource_group=self.aws_resource_group,
            name_filter=name_filter,
            type_filter=type_filter,
            sort_order="delete",
        )
        if aws_resources_to_delete is None or is_empty(aws_resources_to_delete):
            print_info("No AwsResources to delete")
            return True

        # track the total number of AwsResources to delete for validation
        num_resources_to_delete: int = len(aws_resources_to_delete)
        num_resources_deleted: int = 0

        # Print the resources for validation
        print_subheading(f"Deleting {num_resources_to_delete} AwsResources:")
        for rsrc in aws_resources_to_delete:
            print_info(f"  -+-> {rsrc.get_resource_type()}: {rsrc.get_resource_name()}")
        confirm = confirm_yes_no("\nConfirm delete")
        if not confirm:
            print_info("Skipping delete")
            return False

        for resource in aws_resources_to_delete:
            if resource:
                print_info(
                    f"\n-==+==- {resource.get_resource_type()}: {resource.get_resource_name()}"
                )
                # logger.debug(resource)
                try:
                    _resource_deleted = resource.delete(aws_client=self.aws_api_client)
                    if _resource_deleted:
                        num_resources_deleted += 1
                except Exception as e:
                    print_error(
                        f"-==+==--> Resource {resource.get_resource_type()}: {resource.get_resource_name()} could not be deleted."
                    )
                    print_error(e)
                    print_error("Skipping resource deletion, please delete manually...")

        print_info(
            f"\n# Resources deleted: {num_resources_deleted}/{num_resources_to_delete}"
        )
        if num_resources_to_delete == num_resources_deleted:
            return True

        print_error(
            f"Resources deleted: {num_resources_deleted} do not match Resources required: {num_resources_to_delete}"
        )
        return False

    def delete_resources_dry_run(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
    ) -> None:

        env = self.aws_args.env
        print_subheading(
            "AwsResources{}".format(f" for env: {env}" if env is not None else "")
        )
        if self.aws_resource_group is None:
            self.init_resources()
            if self.aws_resource_group is None:
                print_info("No AwsResources available")
                return

        aws_resources_to_delete: Optional[
            List[AwsResourceType]
        ] = filter_and_flatten_aws_resource_group(
            aws_resource_group=self.aws_resource_group,
            name_filter=name_filter,
            type_filter=type_filter,
            sort_order="delete",
        )
        if aws_resources_to_delete is None or is_empty(aws_resources_to_delete):
            print_info("No AwsResources to delete")
            return

        num_resources_to_delete: int = len(aws_resources_to_delete)
        print_info(
            f"\n-*- Shutting down this AwsConfig will delete {num_resources_to_delete} resources:"
        )
        for resource in aws_resources_to_delete:
            print_info(
                f"-==+==- {resource.__class__.__name__}: {resource.get_resource_name()}"
            )

    ######################################################
    ## Patch Resources
    ######################################################

    def patch_resources(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
    ) -> bool:

        logger.debug("-*- Deleting AwsResources")
        if self.aws_resource_group is None:
            self.init_resources()
            if self.aws_resource_group is None:
                print_info("No AwsResources available")
                return False

        aws_resources_to_patch: Optional[
            List[AwsResourceType]
        ] = filter_and_flatten_aws_resource_group(
            aws_resource_group=self.aws_resource_group,
            name_filter=name_filter,
            type_filter=type_filter,
            sort_order="create",
        )
        if aws_resources_to_patch is None or is_empty(aws_resources_to_patch):
            print_info("No AwsResources to patch")
            return True

        # track the total number of AwsResources to patch for validation
        num_resources_to_patch: int = len(aws_resources_to_patch)
        num_resources_patched: int = 0

        # Print the resources for validation
        print_subheading(f"Patching {num_resources_to_patch} AwsResources:")
        for rsrc in aws_resources_to_patch:
            print_info(f"  -+-> {rsrc.get_resource_type()}: {rsrc.get_resource_name()}")
        confirm = confirm_yes_no("\nConfirm patch")
        if not confirm:
            print_info("Skipping patch")
            return False

        for resource in aws_resources_to_patch:
            if resource:
                print_info(
                    f"\n-==+==- {resource.get_resource_type()}: {resource.get_resource_name()}"
                )
                # logger.debug(resource)
                try:
                    _resource_patched = resource.update(aws_client=self.aws_api_client)
                    if _resource_patched:
                        num_resources_patched += 1
                except Exception as e:
                    print_error(
                        f"-==+==--> Resource {resource.get_resource_type()}: {resource.get_resource_name()} could not be patched."
                    )
                    print_error(e)
                    print_error("Skipping resource, please patch manually...")

        print_info(
            f"\n# Resources patched: {num_resources_patched}/{num_resources_to_patch}"
        )
        if num_resources_to_patch == num_resources_patched:
            return True

        print_error(
            f"Resources patched: {num_resources_patched} do not match Resources required: {num_resources_to_patch}"
        )
        return False

    def patch_resources_dry_run(
        self,
        name_filter: Optional[str] = None,
        type_filter: Optional[str] = None,
    ) -> None:

        env = self.aws_args.env
        print_subheading(
            "AwsResources{}".format(f" for env: {env}" if env is not None else "")
        )
        if self.aws_resource_group is None:
            self.init_resources()
            if self.aws_resource_group is None:
                print_info("No AwsResources available")
                return

        aws_resources_to_patch: Optional[
            List[AwsResourceType]
        ] = filter_and_flatten_aws_resource_group(
            aws_resource_group=self.aws_resource_group,
            name_filter=name_filter,
            type_filter=type_filter,
            sort_order="create",
        )
        if aws_resources_to_patch is None or is_empty(aws_resources_to_patch):
            print_info("No AwsResources to patch")
            return

        num_resources_to_patch: int = len(aws_resources_to_patch)
        print_info(
            f"\n-*- Updating this AwsConfig will patch {num_resources_to_patch} resources:"
        )
        for resource in aws_resources_to_patch:
            print_info(
                f"-==+==- {resource.__class__.__name__}: {resource.get_resource_name()}"
            )

    ######################################################
    ## End
    ######################################################



from pathlib import Path
from typing import Optional, Dict
from pydantic import BaseModel

from phidata.infra.docker.resource.group import (
    DockerResourceGroup,
    DockerBuildContext,
)
from phidata.infra.k8s.resource.group import (
    K8sResourceGroup,
    K8sBuildContext,
)
from phidata.utils.log import logger


class PhidataAppArgs(BaseModel):
    name: str
    version: str
    enabled: bool

    # If True, skip resource creation if active resources with the same name exist.
    use_cache: bool = True
    # If True, log extra debug messages
    use_verbose_logs: bool = False

    # Populated during init_resources()
    # These are passed down from the WorkspaceConfig and each PhidataApp has access to them
    # Path to the workspace root directory
    workspace_root_path: Optional[Path] = None
    # Path to the workspace config file
    workspace_config_file_path: Optional[Path] = None

    # Path to important directories relative to the workspace_root
    # These directories are joined to the workspace_root dir
    #   to build paths depending on the environments (local, docker, k8s)
    # defaults are set by WorkspaceConfig.__init__()
    scripts_dir: Optional[str] = None
    storage_dir: Optional[str] = None
    meta_dir: Optional[str] = None
    products_dir: Optional[str] = None
    notebooks_dir: Optional[str] = None
    workspace_config_dir: Optional[str] = None

    # Env vars added to docker env when building PhidataApps
    #   and running workflows
    docker_env: Optional[Dict[str, str]] = None
    # Env vars added to k8s env when building PhidataApps
    #   and running workflows
    k8s_env: Optional[Dict[str, str]] = None

    class Config:
        arbitrary_types_allowed = True


class PhidataApp:
    """Base Class for all PhidataApps"""

    def __init__(self) -> None:
        self.args: Optional[PhidataAppArgs] = None
        self.docker_resource_groups: Optional[Dict[str, DockerResourceGroup]] = None
        self.k8s_resource_groups: Optional[Dict[str, K8sResourceGroup]] = None

    @property
    def name(self) -> str:
        if self.args and self.args.name:
            return self.args.name
        return self.__class__.__name__

    @property
    def version(self) -> Optional[str]:
        return self.args.version if self.args else None

    @property
    def enabled(self) -> bool:
        return self.args.enabled if self.args else False

    @property
    def workspace_root_path(self) -> Optional[Path]:
        return self.args.workspace_root_path if self.args else None

    @workspace_root_path.setter
    def workspace_root_path(self, workspace_root_path: Path) -> None:
        if self.args is not None and workspace_root_path is not None:
            self.args.workspace_root_path = workspace_root_path

    @property
    def workspace_config_file_path(self) -> Optional[Path]:
        return self.args.workspace_config_file_path if self.args else None

    @workspace_config_file_path.setter
    def workspace_config_file_path(self, workspace_config_file_path: Path) -> None:
        if self.args is not None and workspace_config_file_path is not None:
            self.args.workspace_config_file_path = workspace_config_file_path

    @property
    def scripts_dir(self) -> Optional[str]:
        return self.args.scripts_dir if self.args else None

    @scripts_dir.setter
    def scripts_dir(self, scripts_dir: str) -> None:
        if self.args is not None and scripts_dir is not None:
            self.args.scripts_dir = scripts_dir

    @property
    def storage_dir(self) -> Optional[str]:
        return self.args.storage_dir if self.args else None

    @storage_dir.setter
    def storage_dir(self, storage_dir: str) -> None:
        if self.args is not None and storage_dir is not None:
            self.args.storage_dir = storage_dir

    @property
    def meta_dir(self) -> Optional[str]:
        return self.args.meta_dir if self.args else None

    @meta_dir.setter
    def meta_dir(self, meta_dir: str) -> None:
        if self.args is not None and meta_dir is not None:
            self.args.meta_dir = meta_dir

    @property
    def products_dir(self) -> Optional[str]:
        return self.args.products_dir if self.args else None

    @products_dir.setter
    def products_dir(self, products_dir: str) -> None:
        if self.args is not None and products_dir is not None:
            self.args.products_dir = products_dir

    @property
    def notebooks_dir(self) -> Optional[str]:
        return self.args.notebooks_dir if self.args else None

    @notebooks_dir.setter
    def notebooks_dir(self, notebooks_dir: str) -> None:
        if self.args is not None and notebooks_dir is not None:
            self.args.notebooks_dir = notebooks_dir

    @property
    def workspace_config_dir(self) -> Optional[str]:
        return self.args.workspace_config_dir if self.args else None

    @workspace_config_dir.setter
    def workspace_config_dir(self, workspace_config_dir: str) -> None:
        if self.args is not None and workspace_config_dir is not None:
            self.args.workspace_config_dir = workspace_config_dir

    @property
    def docker_env(self) -> Optional[Dict[str, str]]:
        return self.args.docker_env if self.args else None

    @docker_env.setter
    def docker_env(self, docker_env: Dict[str, str]) -> None:
        if self.args is not None and docker_env is not None:
            self.args.docker_env = docker_env

    @property
    def k8s_env(self) -> Optional[Dict[str, str]]:
        return self.args.k8s_env if self.args else None

    @k8s_env.setter
    def k8s_env(self, k8s_env: Dict[str, str]) -> None:
        if self.args is not None and k8s_env is not None:
            self.args.k8s_env = k8s_env

    ######################################################
    ## DockerResourceGroup
    ######################################################

    def init_docker_resource_groups(
        self, docker_build_context: DockerBuildContext
    ) -> None:
        logger.debug(
            f"@init_docker_resource_groups not defined for {self.__class__.__name__}"
        )

    def get_docker_resource_groups(
        self, docker_build_context: DockerBuildContext
    ) -> Optional[Dict[str, DockerResourceGroup]]:
        if self.docker_resource_groups is None:
            self.init_docker_resource_groups(docker_build_context)
        # # Comment out in prod
        # if self.docker_resource_groups:
        #     logger.debug("DockerResourceGroups:")
        #     for rg_name, rg in self.docker_resource_groups.items():
        #         logger.debug(
        #             "{}: {}".format(rg_name, rg.json(exclude_none=True, indent=2))
        #         )
        return self.docker_resource_groups

    ######################################################
    ## K8sResourceGroup
    ######################################################

    def init_k8s_resource_groups(self, k8s_build_context: K8sBuildContext) -> None:
        logger.debug(
            f"@init_docker_resource_groups not defined for {self.__class__.__name__}"
        )

    def get_k8s_resource_groups(
        self, k8s_build_context: K8sBuildContext
    ) -> Optional[Dict[str, K8sResourceGroup]]:
        if self.k8s_resource_groups is None:
            self.init_k8s_resource_groups(k8s_build_context)
        # # Comment out in prod
        # if self.k8s_resource_groups:
        #     logger.debug("K8sResourceGroups:")
        #     for rg_name, rg in self.k8s_resource_groups.items():
        #         logger.debug(
        #             "{}:{}\n{}".format(rg_name, type(rg), rg)
        #         )
        return self.k8s_resource_groups



from setuptools import find_packages, setup

with open("README.md", "r") as rm:
    long_description = rm.read()

version = "0.1.31"

setup(
    name="phidata",
    version=version,
    author="Ashpreet Bedi",
    author_email="ashpreet@phidata.com",
    description="Build data products as code",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://www.phidata.com",
    packages=find_packages(),
    install_requires=[
        "boto3",
        "docker",
        "kubernetes",
        f"phicli=={version}",
        "pydantic",
        "pyyaml",
        "rich",
        "types-PyYAML",
        "typing-extensions",
    ],
    python_requires=">=3.7",
)



from collections import OrderedDict
from pathlib import Path
from typing import Optional, Dict, Any, List, Union
from typing_extensions import Literal

from phidata.app.db import DbApp
from phidata.app.databox.databox import Databox, DataboxArgs
from phidata.infra.docker.resource.container import DockerContainer
from phidata.infra.docker.resource.group import (
    DockerResourceGroup,
    DockerBuildContext,
)
from phidata.infra.k8s.enums.image_pull_policy import ImagePullPolicy
from phidata.infra.k8s.enums.restart_policy import RestartPolicy
from phidata.utils.cli_console import print_error, print_info
from phidata.utils.log import logger

default_devbox_name: str = "devbox"


class DevboxArgs(DataboxArgs):
    install_phidata_dev: bool = False
    phidata_volume_name: str = "devbox-phidata-volume"
    phidata_dir_path: Path = Path.home().joinpath("philab", "phidata")
    phidata_dir_container_path: str = "/phidata"


class Devbox(Databox):
    def __init__(
        self,
        name: str = default_devbox_name,
        version: str = "1",
        enabled: bool = True,
        # Image args,
        image_name: str = "phidata/devbox",
        image_tag: str = "1",
        entrypoint: Optional[Union[str, List]] = None,
        command: Optional[Union[str, List]] = None,
        # Mount the workspace directory on the container,
        mount_workspace: bool = True,
        workspace_volume_name: str = "databox-ws-volume",
        # Path to mount the workspace volume under,
        # This is the parent directory for the workspace on the container,
        # i.e. the ws is mounted as a subdir in this dir,
        # eg: if ws name is: idata, workspace path would be: /usr/local/workspaces/idata,
        workspace_parent_container_path: str = "/usr/local/workspaces",
        # NOTE: On DockerContainers the workspace_root_path is mounted to workspace_dir,
        # because we assume that DockerContainers are running locally on the user's machine,
        # On K8sContainers, we load the workspace_dir from git using a git-sync sidecar container,
        create_git_sync_sidecar: bool = True,
        git_sync_repo: Optional[str] = None,
        git_sync_branch: Optional[str] = None,
        git_sync_wait: int = 1,
        # Install python dependencies using a requirements.txt file,
        install_requirements: bool = True,
        # Path to the requirements.txt file relative to the workspace root,
        requirements_file_path: str = "requirements.txt",
        # Only on DockerContainers, for K8sContainers use IamRole
        # Mount aws config on the container
        mount_aws_config: bool = False,
        # Aws config dir on the host,
        aws_config_path: Path = Path.home().resolve().joinpath(".aws"),
        # Aws config dir on the container,
        aws_config_container_path: str = "/root/.aws",
        # Only on DockerContainers,
        # Mount airflow home from container to host machine,
        # Useful when debugging the airflow conf,
        mount_airflow_home: bool = False,
        # Path to the dir on host machine relative to the workspace root,
        airflow_home_dir: str = "airflow",
        # Path to airflow home on the container,
        airflow_home_container_path: str = "/usr/local/airflow",
        # Configure airflow,
        # If init_airflow = True, this databox initializes airflow and,
        # add env var INIT_AIRFLOW = True which is required by phidata to build dags,
        init_airflow: bool = True,
        # If use_products_as_airflow_dags = True,
        # set the AIRFLOW__CORE__DAGS_FOLDER to the products_dir,
        use_products_as_airflow_dags: bool = True,
        # If use_products_as_airflow_dags = False,
        # set the AIRFLOW__CORE__DAGS_FOLDER to the airflow_dags_path,
        # airflow_dags_path is the directory in the container containing the airflow dags,
        airflow_dags_path: Optional[str] = None,
        # Creates an airflow user with username: test, pass: test,
        create_airflow_test_user: bool = False,
        airflow_executor: Literal[
            "DebugExecutor",
            "LocalExecutor",
            "SequentialExecutor",
            "CeleryExecutor",
            "CeleryKubernetesExecutor",
            "DaskExecutor",
            "KubernetesExecutor",
        ] = "SequentialExecutor",
        logging_level: Literal[
            "DEBUG",
            "INFO",
            "WARNING",
            "ERROR",
            "CRITICAL",
        ] = "INFO",
        # Configure airflow db,
        # If True, initialize the airflow_db on this databox,
        # If None, value is derived from init_airflow i.e. initialize the airflow_db if init_airflow = True,
        #   Locally, airflow_db uses sqllite,
        # If using the databox with an external Airflow db, set init_airflow_db = False,
        init_airflow_db: Optional[bool] = None,
        wait_for_airflow_db: bool = False,
        # Connect to database using DbApp,
        airflow_db_app: Optional[DbApp] = None,
        # Connect to database manually,
        airflow_db_user: Optional[str] = None,
        airflow_db_password: Optional[str] = None,
        airflow_db_schema: Optional[str] = None,
        airflow_db_host: Optional[str] = None,
        airflow_db_port: Optional[str] = None,
        airflow_db_driver: str = "postgresql+psycopg2",
        # Airflow db connections in the format { conn_id: conn_url },
        # converted to env var: AIRFLOW_CONN__conn_id = conn_url,
        db_connections: Optional[Dict] = None,
        # Configure airflow scheduler,
        # Init Airflow scheduler as a daemon process,
        init_airflow_scheduler: bool = False,
        # Configure airflow webserver,
        # Init Airflow webserver when the container starts,
        init_airflow_webserver: bool = False,
        # Open the port if init_airflow_webserver=True,
        # webserver port on the container,
        airflow_webserver_container_port: int = 8080,
        # webserver port on the host machine,
        airflow_webserver_host_port: int = 8080,
        # webserver port name on K8sContainer,
        airflow_webserver_port_name: str = "http",
        # Configure the container,
        container_name: Optional[str] = None,
        image_pull_policy: ImagePullPolicy = ImagePullPolicy.IF_NOT_PRESENT,
        # Only used by the DockerContainer,
        container_detach: bool = True,
        container_auto_remove: bool = True,
        container_remove: bool = True,
        # Add env variables to container env,
        env: Optional[Dict[str, str]] = None,
        # Read env variables from a file in yaml format,
        env_file: Optional[Path] = None,
        # Configure the ConfigMap used for env variables that are not Secret,
        config_map_name: Optional[str] = None,
        # Configure the Secret used for env variables that are Secret,
        secret_name: Optional[str] = None,
        # Read secrets from a file in yaml format,
        secrets_file: Optional[Path] = None,
        # Configure the databox deploy,
        deploy_name: Optional[str] = None,
        pod_name: Optional[str] = None,
        replicas: int = 1,
        pod_node_selector: Optional[Dict[str, str]] = None,
        restart_policy: RestartPolicy = RestartPolicy.ALWAYS,
        termination_grace_period_seconds: Optional[int] = None,
        # Other args,
        load_examples: bool = False,
        print_env_on_load: bool = True,
        # Additional args
        # If True, skip resource creation if active resources with the same name exist.
        use_cache: bool = True,
        # If True, log extra debug messages
        use_verbose_logs: bool = False,
        install_phidata_dev: bool = False,
        phidata_volume_name: str = "devbox-phidata-volume",
        phidata_dir_path: Path = Path.home().joinpath("philab", "phidata"),
        phidata_dir_container_path: str = "/phidata",
    ):
        super().__init__()
        try:
            self.args: DevboxArgs = DevboxArgs(
                name=name,
                version=version,
                enabled=enabled,
                image_name=image_name,
                image_tag=image_tag,
                entrypoint=entrypoint,
                command=command,
                mount_workspace=mount_workspace,
                workspace_volume_name=workspace_volume_name,
                workspace_parent_container_path=workspace_parent_container_path,
                create_git_sync_sidecar=create_git_sync_sidecar,
                git_sync_repo=git_sync_repo,
                git_sync_branch=git_sync_branch,
                git_sync_wait=git_sync_wait,
                install_requirements=install_requirements,
                requirements_file_path=requirements_file_path,
                mount_aws_config=mount_aws_config,
                aws_config_path=aws_config_path,
                aws_config_container_path=aws_config_container_path,
                mount_airflow_home=mount_airflow_home,
                airflow_home_dir=airflow_home_dir,
                airflow_home_container_path=airflow_home_container_path,
                init_airflow=init_airflow,
                use_products_as_airflow_dags=use_products_as_airflow_dags,
                airflow_dags_path=airflow_dags_path,
                create_airflow_test_user=create_airflow_test_user,
                airflow_executor=airflow_executor,
                logging_level=logging_level,
                init_airflow_db=init_airflow_db,
                wait_for_airflow_db=wait_for_airflow_db,
                airflow_db_app=airflow_db_app,
                airflow_db_user=airflow_db_user,
                airflow_db_password=airflow_db_password,
                airflow_db_schema=airflow_db_schema,
                airflow_db_host=airflow_db_host,
                airflow_db_port=airflow_db_port,
                airflow_db_driver=airflow_db_driver,
                db_connections=db_connections,
                init_airflow_scheduler=init_airflow_scheduler,
                init_airflow_webserver=init_airflow_webserver,
                airflow_webserver_container_port=airflow_webserver_container_port,
                airflow_webserver_host_port=airflow_webserver_host_port,
                airflow_webserver_port_name=airflow_webserver_port_name,
                container_name=container_name,
                image_pull_policy=image_pull_policy,
                container_detach=container_detach,
                container_auto_remove=container_auto_remove,
                container_remove=container_remove,
                env=env,
                env_file=env_file,
                config_map_name=config_map_name,
                secret_name=secret_name,
                secrets_file=secrets_file,
                deploy_name=deploy_name,
                pod_name=pod_name,
                replicas=replicas,
                pod_node_selector=pod_node_selector,
                restart_policy=restart_policy,
                termination_grace_period_seconds=termination_grace_period_seconds,
                load_examples=load_examples,
                print_env_on_load=print_env_on_load,
                use_cache=use_cache,
                use_verbose_logs=use_verbose_logs,
                install_phidata_dev=install_phidata_dev,
                phidata_volume_name=phidata_volume_name,
                phidata_dir_path=phidata_dir_path,
                phidata_dir_container_path=phidata_dir_container_path,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    ######################################################
    ## Docker Resources
    ######################################################

    def get_docker_rg(
        self, docker_build_context: DockerBuildContext
    ) -> Optional[DockerResourceGroup]:

        databox_docker_rg = super().get_docker_rg(docker_build_context)
        container_to_update: Optional[DockerContainer] = None
        container_name = self.get_container_name()
        for idx, container in enumerate(databox_docker_rg.containers, start=0):
            if container.name == container_name:
                container_to_update = databox_docker_rg.containers.pop(idx)

        if container_to_update is None:
            print_error("No databox container")
            return None
        # logger.debug(f"Updating container: {container_to_update}")

        # Install phidata in dev mode
        if self.args.install_phidata_dev:
            # Update container environment
            container_env: Dict[str, Any] = {
                "INSTALL_PHIDATA_DEV": str(self.args.install_phidata_dev),
                "PHIDATA_DIR_PATH": self.args.phidata_dir_container_path,
            }
            container_to_update.environment.update(container_env)

            # Create a volume for phidata dev
            if self.args.phidata_dir_path is not None:
                phidata_dir_absolute_path_str = str(self.args.phidata_dir_path)
                logger.debug(f"Mounting: {phidata_dir_absolute_path_str}")
                logger.debug(f"\tto: {self.args.phidata_dir_container_path}")
                container_to_update.volumes[phidata_dir_absolute_path_str] = {
                    "bind": self.args.phidata_dir_container_path,
                    "mode": "rw",
                }
        databox_docker_rg.containers.append(container_to_update)
        return databox_docker_rg

    def init_docker_resource_groups(
        self, docker_build_context: DockerBuildContext
    ) -> None:
        docker_rg = self.get_docker_rg(docker_build_context)
        if docker_rg is not None:
            if self.docker_resource_groups is None:
                self.docker_resource_groups = OrderedDict()
            self.docker_resource_groups[docker_rg.name] = docker_rg



from phidata.utils.enums import ExtendedEnum


class DockerManagerStatus(ExtendedEnum):
    """Enum describing the current status of a DockerManager"""

    # Level 0: The DockerManager has just been created
    PRE_INIT = "PRE_INIT"

    # Level 1: Ready to deploy resources
    WORKER_READY = "WORKER_READY"

    # Level 2: Resources initialized
    RESOURCES_INIT = "RESOURCES_INIT"

    # Level 3: Resources active
    RESOURCES_ACTIVE = "RESOURCES_ACTIVE"

    # Errors
    ERROR = "ERROR"

    def can_create_resources(self) -> bool:
        return self in (
            DockerManagerStatus.WORKER_READY,
            DockerManagerStatus.RESOURCES_INIT,
            DockerManagerStatus.RESOURCES_ACTIVE,
        )

    def can_delete_resources(self) -> bool:
        return self in (
            DockerManagerStatus.WORKER_READY,
            # DockerManagerStatus.RESOURCES_INIT,
            DockerManagerStatus.RESOURCES_ACTIVE,
        )

    def can_get_resources(self) -> bool:
        return self in (
            DockerManagerStatus.WORKER_READY,
            DockerManagerStatus.RESOURCES_INIT,
            DockerManagerStatus.RESOURCES_ACTIVE,
        )



from typing import Dict, Any
import functools

from phidata.utils.env_var import validate_env_vars


def validate_env(env_var_dict: Dict[str, Any]):
    """
    This decorator returns the function it decorates
    IF each env_var in env_var_dict matches its value in the running environment

    Args:
        env_var_dict:

    Returns:

    Notes:
        - Create a decorator with an argument
        https://realpython.com/primer-on-python-decorators/#decorators-with-arguments
    """
    # Return this decorator if env is valid
    def decorator_validate_env(func):
        @functools.wraps(func)
        def wrapper_validate_env(*args, **kwargs):
            value = func(*args, **kwargs)
            return value

        return wrapper_validate_env

    # Return this decorator instead if the env doesnt match
    def decorator_return_none(func):
        @functools.wraps(func)
        def wrapper_return_none(*args, **kwargs):
            return None

        return wrapper_return_none

    if validate_env_vars(env_var_dict):
        return decorator_validate_env
    return decorator_return_none



from datetime import datetime, timedelta
from pathlib import Path
from typing import Optional, Any, Dict, List

from pydantic import BaseModel

from phidata.constants import (
    SCRIPTS_DIR_ENV_VAR,
    STORAGE_DIR_ENV_VAR,
    META_DIR_ENV_VAR,
    PRODUCTS_DIR_ENV_VAR,
    NOTEBOOKS_DIR_ENV_VAR,
    WORKSPACE_CONFIG_DIR_ENV_VAR,
)
from phidata.utils.cli_console import print_info, print_error
from phidata.utils.dttm import dttm_to_dttm_str, days_ago
from phidata.utils.env_var import validate_env_vars
from phidata.utils.log import logger
from phidata.types.context import PathContext, RunContext, AirflowContext


class WorkflowArgs(BaseModel):
    name: str
    version: Optional[str] = None
    enabled: bool = True

    # Optional: airflow task_id for this workflow, use name if not provided
    task_id: Optional[str] = None
    # Optional: airflow dag_id for this workflow, use name if not provided
    dag_id: Optional[str] = None
    # run_context is provided by wf_operator
    run_context: Optional[RunContext] = None
    # path_context is provided using env variables
    path_context: Optional[PathContext] = None

    # Env vars that validate Airflow is active on the containers
    # These vars are useful in gating sections which should only run on
    # remote containers like creating DAGs and tasks
    validate_airflow_env: Dict[str, Any] = {"INIT_AIRFLOW": True}
    airflow_context: Optional[AirflowContext] = None

    class Config:
        arbitrary_types_allowed = True

    @property
    def airflow_active(self) -> bool:
        return validate_env_vars(self.validate_airflow_env)

    @property
    def run_date(self) -> str:
        if self.run_context is not None and self.run_context.run_date is not None:
            return self.run_context.run_date
        if (
            self.airflow_context is not None
            and self.airflow_context.logical_date is not None
        ):
            return dttm_to_dttm_str(
                self.airflow_context.logical_date, dttm_format="%Y-%m-%d"
            )
        return dttm_to_dttm_str(days_ago(), dttm_format="%Y-%m-%d")

    @classmethod
    def from_kwargs(cls, kwargs: Optional[Dict] = None):
        if kwargs is None or not isinstance(kwargs, dict):
            return cls()
        # logger.info(f"Loading {cls.__name__} using kwargs")
        args_object = cls(**kwargs)
        validate_airflow_env = kwargs.get("validate_airflow_env", None)
        if validate_env_vars(validate_airflow_env):
            # logger.info("Creating airflow_context")
            airflow_context = AirflowContext(**kwargs)
            args_object.airflow_context = airflow_context
        return args_object


class Workflow:
    """Base Class for all Workflows"""

    def __init__(self) -> None:
        self.args: Optional[WorkflowArgs] = None

    @property
    def name(self) -> str:
        return self.args.name if self.args else self.__class__.__name__

    @property
    def version(self) -> Optional[str]:
        return self.args.version if self.args else None

    @property
    def enabled(self) -> bool:
        return self.args.enabled if self.args else False

    @property
    def task_id(self) -> str:
        return self.args.task_id if (self.args and self.args.task_id) else self.name

    @task_id.setter
    def task_id(self, task_id: str) -> None:
        if self.args is not None and task_id is not None:
            self.args.task_id = task_id

    @property
    def dag_id(self) -> str:
        return self.args.dag_id if (self.args and self.args.dag_id) else self.name

    @dag_id.setter
    def dag_id(self, dag_id: str) -> None:
        if self.args is not None and dag_id is not None:
            self.args.dag_id = dag_id

    @property
    def run_context(self) -> Optional[RunContext]:
        return self.args.run_context if self.args else None

    @run_context.setter
    def run_context(self, run_context: RunContext) -> None:
        if self.args is not None and run_context is not None:
            self.args.run_context = run_context

    @property
    def path_context(self) -> Optional[PathContext]:
        # Workflow not yet initialized
        if self.args is None:
            return None

        if self.args.path_context is not None:
            # use cached value if available
            return self.args.path_context

        logger.debug(f"--++**++--> Loading PathContext from env")
        self.path_context = PathContext()

        import os

        scripts_dir = os.getenv(SCRIPTS_DIR_ENV_VAR)
        storage_dir = os.getenv(STORAGE_DIR_ENV_VAR)
        meta_dir = os.getenv(META_DIR_ENV_VAR)
        products_dir = os.getenv(PRODUCTS_DIR_ENV_VAR)
        notebooks_dir = os.getenv(NOTEBOOKS_DIR_ENV_VAR)
        workspace_config_dir = os.getenv(WORKSPACE_CONFIG_DIR_ENV_VAR)

        if storage_dir is None:
            print_error(f"{STORAGE_DIR_ENV_VAR} not set")
        if products_dir is None:
            print_error(f"{PRODUCTS_DIR_ENV_VAR} not set")

        try:
            if scripts_dir is not None:
                self.path_context.scripts_dir = Path(scripts_dir)
            if storage_dir is not None:
                self.path_context.storage_dir = Path(storage_dir)
            if meta_dir is not None:
                self.path_context.meta_dir = Path(meta_dir)
            if products_dir is not None:
                self.path_context.products_dir = Path(products_dir)
            if notebooks_dir is not None:
                self.path_context.notebooks_dir = Path(notebooks_dir)
            if workspace_config_dir is not None:
                self.path_context.workspace_config_dir = Path(workspace_config_dir)
        except Exception as e:
            raise
        logger.debug(f"--++**++--> PathContext loaded")
        return self.args.path_context

    @path_context.setter
    def path_context(self, path_context: PathContext) -> None:
        if self.args is not None and path_context is not None:
            self.args.path_context = path_context

    @property
    def validate_airflow_env(self) -> Optional[Dict[str, Any]]:
        return self.args.validate_airflow_env if self.args else None

    @validate_airflow_env.setter
    def validate_airflow_env(self, validate_airflow_env: Dict[str, Any]) -> None:
        if self.args is not None and validate_airflow_env is not None:
            self.args.validate_airflow_env = validate_airflow_env

    @property
    def airflow_active(self) -> bool:
        return validate_env_vars(self.validate_airflow_env)

    ######################################################
    ## Build and validate workflow
    ######################################################

    def build(self) -> bool:
        logger.debug(f"@build not defined for {self.__class__.__name__}")
        return False

    ######################################################
    ## Run workflow
    ######################################################

    def run_in_local_env(self) -> bool:
        """
        Runs a workflow in the local environment where phi wf is called from.

        Returns:
            run_status (bool): True if the run was successful
        """
        logger.debug(f"@run_in_local_env not defined for {self.__class__.__name__}")
        return False

    def run_in_docker_container(
        self, active_container: Any, docker_env: Optional[Dict[str, str]] = None
    ) -> bool:
        """
        Runs a workflow in a docker container.

        Args:
            active_container:
            docker_env:

        Returns:
            run_status (bool): True if the run was successful

        Notes:
            * This function runs in the local environment where phi wf is called from.
            But executes `airflow` commands in the docker container to run the workflow
            * For the airflow tasks to be available, they need to be added to the workflow DAG
            using add_airflow_tasks_to_dag()
        """
        logger.debug(
            f"@run_in_docker_container not defined for {self.__class__.__name__}"
        )
        return False

    def run_in_k8s_container(
        self,
        pod: Any,
        k8s_api_client: Any,
        container_name: Optional[str] = None,
        k8s_env: Optional[Dict[str, str]] = None,
    ) -> bool:
        logger.debug(f"@run_in_k8s_container not defined for {self.__class__.__name__}")
        return False

    ######################################################
    ## Airflow functions
    ######################################################

    def add_airflow_tasks_to_dag(self, dag: Any) -> bool:
        """
        Add tasks to the airflow DAG.

        Args:
            dag:

        Returns:

        Notes:
            * This function is called as part of the create_airflow_dag() function
        """
        logger.debug(
            f"@add_airflow_tasks_to_dag not defined for {self.__class__.__name__}"
        )
        return False

    def create_airflow_dag(
        self,
        owner: Optional[str] = "airflow",
        depends_on_past: Optional[bool] = False,
        # The description for the DAG to e.g. be shown on the webserver
        description: Optional[str] = None,
        # Defines how often that DAG runs, this
        #  timedelta object gets added to your latest task instance's
        #  execution_date to figure out the next schedule
        # Default: daily
        schedule_interval: timedelta = timedelta(days=1),
        # The timestamp from which the scheduler will
        #  attempt to backfill
        start_date: Optional[datetime] = None,
        # if start_date is not provided, use start_days_ago
        start_days_ago: int = 2,
        # A date beyond which your DAG won't run, leave to None
        #  for open ended scheduling
        end_date: Optional[datetime] = None,
        # a dictionary of macros that will be exposed
        #  in your jinja templates. For example, passing ``dict(foo='bar')``
        #  to this argument allows you to ``{{ foo }}`` in all jinja
        #  templates related to this DAG. Note that you can pass any
        #  type of object here.
        user_defined_macros: Optional[Dict] = None,
        # a dictionary of filters that will be exposed
        #  in your jinja templates. For example, passing
        #  ``dict(hello=lambda name: 'Hello %s' % name)`` to this argument allows
        #  you to ``{{ 'world' | hello }}`` in all jinja templates related to
        #  this DAG.
        user_defined_filters: Optional[Dict] = None,
        # A dictionary of default parameters to be used
        #  as constructor keyword parameters when initialising operators.
        #  Note that operators have the same hook, and precede those defined
        #  here, meaning that if your dict contains `'depends_on_past': True`
        #  here and `'depends_on_past': False` in the operator's call
        #  `default_args`, the actual value will be `False`.
        default_args: Optional[Dict] = None,
        # the number of task instances allowed to run concurrently
        concurrency: Optional[int] = None,
        # maximum number of active DAG runs, beyond this
        #  number of DAG runs in a running state, the scheduler won't create
        #  new active DAG runs
        max_active_runs: int = 8,
        doc_md: Optional[str] = None,
        # a dictionary of DAG level parameters that are made
        # accessible in templates, namespaced under `params`. These
        # params can be overridden at the task level.
        params: Optional[Dict] = None,
        is_paused_upon_creation: Optional[bool] = None,
        jinja_environment_kwargs: Optional[Dict] = None,
        tags: Optional[List[str]] = None,
    ) -> Any:
        """
        Used to create an Airflow DAG for independent workflows.
        It is preferred to create 1 DAG for each DataProduct but not all pipelines
        have DataProducts, some just have 1 Workflow.
        """

        # Skips DAG creation on local machines
        if not self.airflow_active:
            return None

        from airflow import DAG
        from airflow.utils.dates import days_ago

        operator_default_args = default_args
        if operator_default_args is None:
            operator_default_args = {}
        operator_default_args.update(
            {
                "owner": owner,
                "depends_on_past": depends_on_past,
            }
        )
        if start_date is None:
            start_date = days_ago(start_days_ago)

        dag_id = self.dag_id
        if dag_id is None:
            print_error("Workflow dag_id unavailable")
            return False
        dag = DAG(
            dag_id=dag_id,
            description=description,
            schedule_interval=schedule_interval,
            start_date=start_date,
            end_date=end_date,
            user_defined_macros=user_defined_macros,
            user_defined_filters=user_defined_filters,
            default_args=operator_default_args,
            concurrency=concurrency,
            max_active_runs=max_active_runs,
            doc_md=doc_md,
            params=params,
            is_paused_upon_creation=is_paused_upon_creation,
            jinja_environment_kwargs=jinja_environment_kwargs,
            tags=tags,
        )
        add_task_success = self.add_airflow_tasks_to_dag(dag)
        if not add_task_success:
            print_error(f"Tasks for workflow {self.name} could not be added")
        logger.debug(f"Airflow dag {dag_id} created")
        return dag
